"""Tests for coordinate manager."""
from __future__ import annotations

import numpy as np
import pytest
from pydantic import ValidationError
from rich.text import Text

import dascore as dc
import dascore.proc.coords
from dascore import to_datetime64
from dascore.core.coordmanager import (
    CoordManager,
    get_coord_manager,
    merge_coord_managers,
)
from dascore.core.coords import (
    BaseCoord,
    CoordArray,
    CoordMonotonicArray,
    CoordRange,
    get_coord,
)
from dascore.exceptions import (
    CoordDataError,
    CoordError,
    CoordMergeError,
    CoordSortError,
    ParameterError,
)
from dascore.units import get_quantity
from dascore.utils.misc import all_close, get_middle_value, register_func

COORD_MANAGERS = []

COORDS = {
    "time": to_datetime64(np.arange(10, 100, 10)),
    "distance": get_coord(values=np.arange(0, 1_000, 10)),
}
DIMS = ("time", "distance")


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def basic_coord_manager():
    """The simplest coord manager."""
    return get_coord_manager(COORDS, DIMS)


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_manager_with_units(basic_coord_manager):
    """The simplest coord manager."""
    return basic_coord_manager.set_units(time="s", distance="m")


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def basic_degenerate_coord_manager(basic_coord_manager):
    """A degenerate coord manager on time axis."""
    time_coord = basic_coord_manager.coord_map["time"]
    degenerate = time_coord.empty()
    return basic_coord_manager.update_coords(time=degenerate)


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_manager_multidim() -> CoordManager:
    """The simplest coord manager with several coords added."""
    COORDS = {
        "time": to_datetime64(np.arange(10, 110, 10)),
        "distance": get_coord(values=np.arange(0, 1000, 10)),
        "quality": (("time", "distance"), np.ones((10, 100))),
        "latitude": ("distance", np.random.rand(100)),
    }
    DIMS = ("time", "distance")

    return get_coord_manager(COORDS, DIMS)


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_manager_degenerate_time(coord_manager_multidim) -> CoordManager:
    """A coordinate manager with degenerate (length 1) time array."""
    new_time = to_datetime64(["2017-09-18T01:00:01"])
    out = coord_manager_multidim.update_coords(time=new_time)
    return out


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_manager_wacky_dims() -> CoordManager:
    """A coordinate manager with non evenly sampled dims."""
    patch = dc.get_example_patch("wacky_dim_coords_patch")
    return patch.coords


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_dt_small_diff(memory_spool_small_dt_differences):
    """A list of coordinate managers with differences in dt merged."""
    spool = memory_spool_small_dt_differences
    coords = [x.coords for x in spool]
    out = merge_coord_managers(coords, dim="time")
    return out


@pytest.fixture(scope="class")
@register_func(COORD_MANAGERS)
def coord_non_associated_coord(basic_coord_manager):
    """A cm with coordinates that are not associated with a dimension."""
    new = basic_coord_manager.update_coords(
        bob=(None, np.arange(10)),
        bill=((), np.arange(100)),
    )
    return new


@pytest.fixture(scope="class", params=COORD_MANAGERS)
def coord_manager(request) -> CoordManager:
    """Meta fixture for aggregating coordinates."""
    return request.getfixturevalue(request.param)


class TestGetCoordManager:
    """Test suite for `get_coord_manger` helper function."""

    def test_coords_and_attrs_raise(self, random_patch):
        """Ensure using coords and attrs raises."""
        msg = "Cannot use both attrs and coords"
        coords, attrs = random_patch.coords, random_patch.attrs
        with pytest.raises(ParameterError, match=msg):
            get_coord_manager(coords=coords, attrs=attrs)

    def test_coords_from_attrs(self, random_patch):
        """Ensure we can get coordinates from patch attrs."""
        attrs = random_patch.attrs
        cm = get_coord_manager(attrs=attrs)
        assert "time" in cm.coord_map


class TestBasicCoordManager:
    """Ensure basic things work with coord managers."""

    def test_init(self, coord_manager):
        """Ensure values can be init'ed."""
        assert isinstance(coord_manager, CoordManager)

    def test_to_dict(self, coord_manager):
        """CoordManager should be convertible to dict."""
        c_dict = dict(coord_manager)
        expected = {x: coord_manager.get_array(x) for x in coord_manager.coord_map}
        assert set(expected) == set(c_dict)
        for key in set(expected):
            assert np.all(expected[key] == c_dict[key])

    def test_membership(self, coord_manager):
        """Coord membership should work for coord names."""
        coords = list(coord_manager.coord_map)
        for name in coords:
            assert name in coords

    def test_empty(self):
        """And empty coord manager should be possible."""
        coord = get_coord_manager()
        assert isinstance(coord, CoordManager)
        assert dict(coord) == {}
        # shape should be the same as an empty array.
        assert coord.shape == np.array([]).shape

    def test_str(self, coord_manager):
        """Tests the str output for coord manager."""
        out = str(coord_manager)
        assert isinstance(out, str)
        assert len(out)

    def test_rich(self, coord_manager):
        """Tests the str output for coord manager."""
        out = coord_manager.__rich__()
        assert isinstance(out, Text)
        assert len(out)

    def test_cant_assign_new_coord_inplace(self, basic_coord_manager):
        """The mappings inside the coord manager should be immutable."""
        cm = basic_coord_manager
        expected_str = "does not support item assignment"
        # cant add new coord
        with pytest.raises(TypeError, match=expected_str):
            cm["bob"] = 10
        # cant modify existing coord
        with pytest.raises(TypeError, match=expected_str):
            cm[cm.dims[0]] = cm.get_array(cm.dims[0])

    def test_cant_modify_dim_map(self, basic_coord_manager):
        """Ensure the dim map is immutable."""
        expected_str = "does not support item assignment"
        dim_map = basic_coord_manager.dim_map
        # test new key
        with pytest.raises(TypeError, match=expected_str):
            dim_map["bob"] = 10
        # test existing key
        with pytest.raises(TypeError, match=expected_str):
            dim_map[basic_coord_manager.dims[0]] = 10

    def test_cant_modify_coord_map(self, basic_coord_manager):
        """Ensure the dim map is immutable."""
        expected_str = "does not support item assignment"
        coord_map = basic_coord_manager.coord_map
        # test new key
        with pytest.raises(TypeError, match=expected_str):
            coord_map["bob"] = 10
        # test existing key
        with pytest.raises(TypeError, match=expected_str):
            coord_map[basic_coord_manager.dims[0]] = 10

    def test_init_with_coord_manager(self, basic_coord_manager):
        """Ensure initing coord manager works with a single coord manager."""
        out = get_coord_manager(basic_coord_manager)
        assert out == basic_coord_manager

    def test_init_with_cm_and_dims(self, basic_coord_manager):
        """Ensure cm can be init'ed with coord manager and dims."""
        out = get_coord_manager(basic_coord_manager, dims=basic_coord_manager.dims)
        assert out == basic_coord_manager

    def test_init_list_range(self):
        """Ensure the coord manager can be init'ed with a list and no dims."""
        input_dict = {
            "time": [10, 11, 12],
            "distance": ("distance", range(100)),
            "space": range(100),
        }
        out = get_coord_manager(input_dict, dims=list(input_dict))
        assert set(out.dims) == set(input_dict)

    def test_bad_datashape_raises(self, basic_coord_manager):
        """Ensure a bad datashape raises."""
        match = "match the coordinate manager shape"
        cm = basic_coord_manager
        data = np.ones((cm.shape[0], cm.shape[1] - 1))
        with pytest.raises(CoordDataError, match=match):
            cm.validate_data(data)

    def test_to_summary_dict(self, coord_manager):
        """Ensure coord managers can be converted to summary dicts."""
        sum_dict = coord_manager.to_summary_dict()
        assert set(sum_dict) == set(coord_manager.coord_map)

    def test_size(self, coord_manager):
        """Ensure all coord managers have a size."""
        assert isinstance(coord_manager.size, int | np.int_)

    def test_min(self, basic_coord_manager):
        """Ensure we can git min value."""
        expected = np.min(basic_coord_manager.time.data).astype(np.int64)
        got = basic_coord_manager.min("time").astype(np.int64)
        assert np.isclose(got, expected)

    def test_max(self, basic_coord_manager):
        """Ensure we can git max value."""
        expected = np.max(basic_coord_manager.time.data).astype(np.int64)
        got = basic_coord_manager.max("time").astype(np.int64)
        assert np.isclose(got, expected)

    def test_step(self, basic_coord_manager):
        """Ensure we can git min value."""
        expected = basic_coord_manager.time.step
        assert basic_coord_manager.step("time") == expected

    def test_getattr(self, basic_coord_manager):
        """Ensure getattr returns coordinate."""
        for dim in basic_coord_manager.dims:
            coord = getattr(basic_coord_manager, dim)
            assert coord == basic_coord_manager.coord_map[dim]

    def test_get_item_warning(self, basic_coord_manager):
        """Ensure get item emits a warning."""
        msg = "returns a numpy array"
        with pytest.warns(UserWarning, match=msg):
            _ = basic_coord_manager["time"]


class TestCoordManagerInputs:
    """Tests for coordinates management."""

    def test_simple_inputs(self):
        """Simplest input case."""
        out = get_coord_manager(COORDS, DIMS)
        assert isinstance(out, CoordManager)

    def test_additional_coords(self):
        """Ensure a additional (non-dimensional) coords work."""
        coords = dict(COORDS)
        lats = np.random.rand(len(COORDS["distance"]))
        coords["latitude"] = ("distance", lats)
        out = get_coord_manager(coords, DIMS)
        assert isinstance(out.coord_map["latitude"], BaseCoord)

    def test_str(self, basic_coord_manager):
        """Ensure a custom (readable) str is returned."""
        coord_str = str(basic_coord_manager)
        assert isinstance(coord_str, str)

    def test_bad_coords(self):
        """Ensure specifying a bad coordinate raises."""
        coords = dict(COORDS)
        coords["bill"] = np.arange(10, 100, 10)
        with pytest.raises(CoordError, match="not named the same as dimension"):
            get_coord_manager(coords, DIMS)

    def test_nested_coord_too_long(self):
        """Nested coordinates that are gt 2 should fail."""
        coords = dict(COORDS)
        coords["time"] = ("time", to_datetime64(np.arange(10, 100, 10)), "space")
        with pytest.raises(CoordError, match="must be length two"):
            get_coord_manager(coords, DIMS)

    def test_invalid_dimensions(self):
        """Nested coordinates must specify valid dimensions."""
        coords = dict(COORDS)
        coords["time"] = ("bob", to_datetime64(np.arange(10, 100, 10)))
        with pytest.raises(CoordError, match="invalid dimension"):
            get_coord_manager(coords, DIMS)

    def test_missing_coordinates(self):
        """If all dims don't have coords an error should be raised."""
        coords = dict(COORDS)
        coords.pop("distance")
        msg = "All dimensions must have coordinates"
        with pytest.raises(ValidationError, match=msg):
            get_coord_manager(coords, DIMS)

    def test_secondary_coord_bad_lengths(self):
        """Ensure when coordinates don't line up an error is raised."""
        coords = dict(COORDS)
        coords["bad"] = ("time", np.ones(len(coords["time"]) - 1))
        with pytest.raises(ValidationError, match="does not match the dimension"):
            get_coord_manager(coords, DIMS)

    def test_dim_coord_wrong_dim(self):
        """It shouldn't be possible to init a dimension coord with the wrong dim."""
        coords = {
            "time": ((), np.array([1, 2, 3])),
            "distance": ("distance", np.array([1, 2, 3])),
        }
        dims = "time", "distance"
        with pytest.raises(ValidationError):
            get_coord_manager(coords, dims=dims)

    def test_mappings_immutable(self, coord_manager):
        """Ensure the mappings are immutable."""
        with pytest.raises(Exception):
            coord_manager.coord_map["bob"] = 2


class TestCoordManagerWithAttrs:
    """Tests for initing coord managing with attribute dict."""

    def test_missing_dim(self):
        """Coord manager should be able to pull missing info from attributes."""
        attrs = dict(distance_min=1, distance_max=100, distance_step=10)
        new = get_coord_manager(None, ("distance",), attrs=attrs)
        assert "distance" in new.coord_map


class TestDrop:
    """Tests for dropping coords with coord manager."""

    def test_drop(self, coord_manager_multidim):
        """Ensure coordinates can be dropped."""
        dim = "distance"
        coords, _ = coord_manager_multidim.drop_coords(dim)
        assert dim not in coords.dims
        for _name, dims in coords.dim_map.items():
            assert dim not in dims

    def test_drop_doesnt_have_coord(self, coord_manager_multidim):
        """Trying to drop a dim that doesnt exist should just return."""
        out, _ = coord_manager_multidim.drop_coords("bob")
        assert out == coord_manager_multidim

    def test_trims_array(self, coord_manager_multidim):
        """Trying to drop a dim that doesnt exist should just return."""
        array = np.ones(coord_manager_multidim.shape)
        axis = coord_manager_multidim.dims.index("time")
        cm, new_array = coord_manager_multidim.drop_coords("time", array=array)
        assert new_array.shape[axis] == 0

    def test_drop_non_dim_coord(self, coord_manager_multidim):
        """Dropping a non-dim coord should not affect shape/dimensions."""
        cm = coord_manager_multidim
        array = np.ones(cm.shape)
        coords_to_drop = set(cm.coord_map) - set(cm.dims)
        for coord in coords_to_drop:
            cm_new, array_new = cm.drop_coords(coord, array=array)
            # array should not have changed.
            assert array.shape == array_new.shape
            assert np.all(np.equal(array, array_new))
            assert coord not in set(cm_new.coord_map)


class TestSelect:
    """Tests for filtering coordinates."""

    # index ranges to compare with normal select.
    slice_inds = (
        (None, 5),
        (..., 5),
        (1, 5),
        (1, None),
        (1, ...),
        (..., ...),
        (None, None),
        (3, 4),
        (-4, -1),
        (1, -1),
    )
    # single index values to test
    inds = (
        0,
        1,
        -1,
        -2,
    )

    def test_2d_coord_raises(self, coord_manager_multidim):
        """Select shouldn't work on 2D coordinates."""
        with pytest.raises(CoordError, match="Only 1 dimensional"):
            coord_manager_multidim.select(quality=(1, 2))

    def test_select_coord_dim(self, basic_coord_manager):
        """Simple test for filtering dimension coord."""
        new, _ = basic_coord_manager.select(distance=(100, 400))
        dist_ind = basic_coord_manager.dims.index("distance")
        assert new.shape[dist_ind] < basic_coord_manager.shape[dist_ind]

    def test_filter_array(self, basic_coord_manager):
        """Ensure an array can be filtered."""
        data = np.ones(basic_coord_manager.shape)
        new, trim = basic_coord_manager.select(distance=(100, 400), array=data)
        assert trim.shape == trim.shape

    def test_select_emptying_dim(self, basic_coord_manager):
        """Selecting a range outside of dim should empty the manager."""
        data = np.ones(basic_coord_manager.shape)
        cm, trim = basic_coord_manager.select(distance=(-100, -10), array=data)
        assert trim.shape[cm.dims.index("distance")] == 0
        assert "distance" in cm.dims
        assert len(cm.get_array("distance")) == 0
        assert len(cm.coord_map["distance"]) == 0

    def test_select_trims_associated_coord_1(self, coord_manager_multidim):
        """Ensure trimming a dimension also trims associated coordinate."""
        cm = coord_manager_multidim
        coord_to_trim = "distance"
        distance = cm.get_array(coord_to_trim)
        out, _ = cm.select(distance=(distance[1], distance[-2]))
        # ensure all attrs with "distance" have been trimmed.
        expected_len = len(out.coord_map[coord_to_trim])
        for name in cm.dim_to_coord_map[coord_to_trim]:
            coord = out.coord_map[name]
            axis = cm.dim_map[name].index(coord_to_trim)
            assert coord.shape[axis] == expected_len

    def test_select_trims_associated_coords_2(self, coord_manager_multidim):
        """Same as test #1, but now we check for trimming non-dimension coord."""
        cm = coord_manager_multidim
        coord_to_trim = "latitude"
        array = cm.get_array(coord_to_trim)
        out, _ = cm.select(**{coord_to_trim: (array[1], array[-2])})
        dim = cm.dim_map[coord_to_trim][0]
        # ensure all attrs with shared dim have been trimmed.
        expected_len = len(out.coord_map[coord_to_trim])
        for name in cm.dim_to_coord_map[dim]:
            coord = out.coord_map[name]
            axis = cm.dim_map[name].index(dim)
            assert coord.shape[axis] == expected_len

    def test_select_handles_non_dim_kwargs(self, basic_coord_manager):
        """The coord manager should handle (supress) non dim keyword args."""
        ar = np.ones(basic_coord_manager.shape)
        out, new = basic_coord_manager.select(bob=(10, 20), array=ar)
        assert new.shape == ar.shape
        assert out == basic_coord_manager

    @pytest.mark.parametrize("slice_range", slice_inds)
    def test_compare_to_select(self, basic_coord_manager, slice_range):
        """Ensure select with and without samples behaves the same with equiv. data."""
        cm = basic_coord_manager
        for name, coord in cm.coord_map.items():
            ind_tuple = slice_range
            ind_1, ind_2 = slice_range
            val1 = coord[ind_1] if isinstance(ind_1, int) else ind_1
            val2 = coord[ind_2 - 1] if isinstance(ind_2, int) else ind_2
            value_tuple = (val1, val2)
            # first, just check coords are equal
            out1 = coord.select(value_tuple)[0]
            out2 = coord[slice(*ind_tuple)]
            if not out1 == out2:
                assert all_close(out1.values, out2.values)
            # then check that the whole coord_manager are equal
            cm1 = cm.select(**{name: value_tuple})
            cm2 = cm.select(**{name: ind_tuple}, samples=True)
            assert cm1 == cm2

    @pytest.mark.parametrize("index", inds)
    def test_single_values(self, basic_coord_manager, index):
        """
        Single values should be treated like slice(val, val+1)
        as not to collapse the dimensions when samples=True.
        """
        cm = basic_coord_manager
        data = np.empty(cm.shape)
        for dim in basic_coord_manager.dims:
            kwargs = {dim: index}
            out1, new_data = cm.select(array=data, samples=True, **kwargs)
            dim_ind = cm.dims.index(dim)
            # now the array should have a len(1) in the selected dimension.
            assert out1.shape[dim_ind] == new_data.shape[dim_ind] == 1
            new_value = out1.coord_map[dim].values[0]
            expected_value = cm.get_array(dim)[index]
            all_close(new_value, expected_value)

    def test_trim_related_coords(self, coord_manager_multidim):
        """Ensure trim also trims related dimensions."""
        cm = coord_manager_multidim
        data = np.empty(cm.shape)
        out, new_data = cm.select(array=data, time=slice(2, 4), samples=True)
        for name, coord in out.coord_map.items():
            dims = cm.dim_map[name]
            if "time" not in dims:
                continue
            time_id = dims.index("time")
            assert coord.shape[time_id] == 2

    def test_samples_slice(self, coord_manager):
        """Ensure we can select when samples=True using ... or None."""
        new, _ = coord_manager.select(time=..., samples=True)
        assert new == coord_manager


class TestEquals:
    """Tests for coord manager equality."""

    def test_basic_equals(self, coord_manager):
        """All coord managers should equal themselves."""
        assert coord_manager == coord_manager

    def test_unequal_float_coords(self, coord_manager_multidim):
        """Ensure if coordinates are not equal false is returned."""
        coord = coord_manager_multidim.coord_map["latitude"]
        new = get_coord(values=coord.values + 10)
        args = dict(latitude=("distance", new))
        new_coord = coord_manager_multidim.update_coords(**args)
        assert new_coord != coord_manager_multidim

    def test_unequal_wrong_type(self, basic_coord_manager):
        """Non-coord managers should not be considered equal."""
        cm = basic_coord_manager
        assert cm != 10
        assert cm != "bob"
        assert cm != {1: 2, 2: 2}
        assert cm != cm.coord_map["distance"]


class TestTranspose:
    """Test suite for transposing dimensions."""

    @pytest.fixture(scope="class")
    def many_dims_cm(self):
        """Create a coordmanager with many dimensions."""
        dims = ("one", "two", "three", "four")
        coords = {x: np.arange(10) for x in dims}
        many_dims = get_coord_manager(coords, dims=dims)
        return many_dims

    def test_simple_transpose(self, basic_coord_manager):
        """Ensure the coord manager can be transposed."""
        dims = basic_coord_manager.dims
        new_dims = dims[::-1]
        tran = basic_coord_manager.transpose(*new_dims)
        assert tran.dims == new_dims
        assert tran.shape != basic_coord_manager.shape
        assert tran.shape == basic_coord_manager.shape[::-1]

    def test_empty_ellipses(self, many_dims_cm):
        """Empty transpose should just reverse order."""
        out = many_dims_cm.transpose()
        assert out.dims == many_dims_cm.dims[::-1]

    def test_ellipses_at_start(self, many_dims_cm):
        """Ensure ellipses works to just stick things at start/end."""
        new = many_dims_cm.transpose(..., "one", "two")
        assert new.dims == ("three", "four", "one", "two")

    def test_ellipses_at_end(self, many_dims_cm):
        """Ensure ellipses works to just stick things at start/end."""
        new = many_dims_cm.transpose("four", "two", ...)
        assert new.dims == ("four", "two", "one", "three")

    def test_ellipses_in_middle(self, many_dims_cm):
        """Ensure ellipses works to just stick things at start/end."""
        new = many_dims_cm.transpose("four", ..., "one")
        assert new.dims == ("four", "two", "three", "one")

    def test_duplicate_raises(self, many_dims_cm):
        """... can only be used once, and repeat dimensions must raise."""
        with pytest.raises(ParameterError, match="duplicate dimensions"):
            many_dims_cm.transpose(..., "two", ...)
        with pytest.raises(ParameterError, match="duplicate dimensions"):
            many_dims_cm.transpose("one", "one", "two", "three", "four")

    def test_raises_no_ellipses_missing_dim(self, many_dims_cm):
        """When ... is not used all dims must appear in args."""
        with pytest.raises(ParameterError, match="specify all dimensions"):
            many_dims_cm.transpose("one", "three", "four")
        with pytest.raises(ParameterError, match="specify all dimensions"):
            many_dims_cm.transpose("one")


class TestRenameDims:
    """Test case for renaming dimensions."""

    def test_rename_dims(self, basic_coord_manager):
        """Ensure dimensions can be renamed."""
        rename_map = {x: x[:2] for x in basic_coord_manager.dims}
        out = basic_coord_manager.rename_coord(**rename_map)
        assert set(out.dims) == set(rename_map.values())

    def test_rename_extra_coords_kept(self, coord_manager_multidim):
        """Ensure the extra dims are not dropped when a coord is renamed."""
        cm = coord_manager_multidim
        # we should be able to rename distance to dist and
        # distance just gets swapped for dist
        out = cm.rename_coord(distance="dist")
        assert "dist" in out.dim_map
        assert "dist" in out.coord_map
        assert "dist" in out.dim_map["latitude"]

    def test_rename_same_dims_extra_coords(self, coord_manager_multidim):
        """Ensure renaming dims the same doesn't mess with multidims."""
        cm = coord_manager_multidim
        dim_map = {x: x for x in cm.dims}
        out = cm.rename_coord(**dim_map)
        assert set(out.dim_map) == set(cm.dim_map)


class TestUpdateFromAttrs:
    """Tests to ensure updating attrs can update coordinates."""

    def test_update_min(self, basic_coord_manager):
        """Ensure min time in attrs updates appropriate coord."""
        for dim in basic_coord_manager.dims:
            coord = basic_coord_manager.coord_map[dim]
            attrs = {f"{dim}_max": coord.min()}
            new, _ = basic_coord_manager.update_from_attrs(attrs)
            new_coord = new.coord_map[dim]
            assert len(new_coord) == len(coord)
            assert new_coord.max() == coord.min()

    def test_update_max(self, basic_coord_manager):
        """Ensure max time in attrs updates appropriate coord."""
        for dim in basic_coord_manager.dims:
            coord = basic_coord_manager.coord_map[dim]
            attrs = {f"{dim}_min": coord.max()}
            dist = coord.max() - coord.min()
            new, _ = basic_coord_manager.update_from_attrs(attrs)
            new_coord = new.coord_map[dim]
            new_dist = new_coord.max() - new_coord.min()
            assert dist == new_dist
            assert len(new_coord) == len(coord)
            assert new_coord.min() == coord.max()

    def test_update_step(self, basic_coord_manager):
        """Ensure the step can be updated which changes endtime."""
        for dim in basic_coord_manager.dims:
            coord = basic_coord_manager.coord_map[dim]
            attrs = {f"{dim}_step": coord.step * 10}
            dist = coord.max() - coord.min()
            new, _ = basic_coord_manager.update_from_attrs(attrs)
            new_coord = new.coord_map[dim]
            new_dist = new_coord.max() - new_coord.min()
            assert (dist * 10) == new_dist
            assert len(new_coord) == len(coord)
            assert new_coord.min() == coord.min()

    def test_attrs_as_dict(self, basic_coord_manager):
        """Ensure the attrs returned has coords attached."""
        coord = basic_coord_manager.coord_map["time"]
        attrs = {"time_max": coord.min()}
        cm, attrs = basic_coord_manager.update_from_attrs(attrs)
        assert attrs.coords == cm.to_summary_dict()
        assert attrs.dim_tuple == cm.dims

    def test_attrs_as_patch_attr(self, basic_coord_manager):
        """Ensure this also works when attrs is a patch attr."""
        attrs = dc.PatchAttrs(time_min=to_datetime64("2022-01-01"))
        cm, new_attrs = basic_coord_manager.update_from_attrs(attrs)
        assert new_attrs.coords == cm.to_summary_dict()
        assert new_attrs.dim_tuple == cm.dims

    def test_consistent_attrs_leaves_coords_unchanged(self, random_patch):
        """Attrs which are already consistent should leave coord unchanged."""
        attrs, coords = random_patch.attrs, random_patch.coords
        new_coords, new_attrs = coords.update_from_attrs(attrs)
        assert new_coords == coords


class TestUpdateCoords:
    """Tests for updating coordinates."""

    def test_simple(self, basic_coord_manager):
        """Ensure coordinates can be updated (replaced)."""
        new_time = basic_coord_manager.get_array("time") + np.timedelta64(1, "s")
        out = basic_coord_manager.update_coords(time=new_time)
        assert np.all(np.equal(out.get_array("time"), new_time))

    def test_extra_coords_kept(self, coord_manager_multidim):
        """Ensure extra coordinates are kept."""
        cm = coord_manager_multidim
        new_time = cm.get_array("time") + np.timedelta64(1, "s")
        out = cm.update_coords(time=new_time)
        assert set(out.coord_map) == set(coord_manager_multidim.coord_map)
        assert set(out.coord_map) == set(coord_manager_multidim.coord_map)

    def test_size_change(self, basic_coord_manager):
        """Ensure sizes of dimensions can be changed."""
        new_time = basic_coord_manager.get_array("time")[:10]
        out = basic_coord_manager.update_coords(time=new_time)
        assert np.all(np.equal(out.get_array("time"), new_time))

    def test_size_change_drops_old_coords(self, coord_manager_multidim):
        """When the size changes on multidim, dims should be dropped."""
        cm = coord_manager_multidim
        new_dist = cm.get_array("distance")[:10]
        dropped_coords = set(cm.coord_map) - set(cm.dims)
        out = cm.update_coords(distance=new_dist)
        assert dropped_coords.isdisjoint(set(out.coord_map))

    def test_update_degenerate(self, coord_manager):
        """Tests for updating coord with degenerate coordinates."""
        cm = coord_manager
        out = cm.update_coords(time=cm.coord_map["time"].empty())
        assert out.shape[out.dims.index("time")] == 0
        assert len(out.coord_map["time"]) == 0

    def test_update_degenerate_dim_multicoord(self, coord_manager_multidim):
        """
        If one dim is degenerate all associated coords should be
        dropped.
        """
        cm = coord_manager_multidim
        degen_time = cm.coord_map["time"].empty()
        new = cm.update_coords(time=degen_time)
        assert new != cm
        # any coords with time as a dim (but not time itself) should be gone.
        has_time = [i for i, v in cm.dim_map.items() if ("time" in v and i != "time")]
        assert set(has_time).isdisjoint(set(new.coord_map))

    def test_update_none_drops(self, basic_coord_manager):
        """Ensure when passing coord=None the coord is dropped."""
        cm1 = basic_coord_manager.update_coords(time=None)
        cm2, _ = basic_coord_manager.drop_coords("time")
        assert cm1 == cm2

    def test_update_only_start(self, basic_coord_manager):
        """Ensure start_coord can be used to update."""
        time1 = basic_coord_manager.coord_map["time"]
        new_start = time1.max()
        cm = basic_coord_manager.update_coords(time_min=new_start)
        time2 = cm.coord_map["time"]
        assert time2.min() == new_start

    def test_dissociate(self, basic_coord_manager):
        """Ensure coordinates can be dissociated with update_coords."""
        time = basic_coord_manager.coord_map["time"]
        new_time = time.values + dc.to_timedelta64(1)
        new = basic_coord_manager.update_coords(new_time=("time", new_time))
        assert "new_time" in new.dim_map and "new_time" in new.coord_map
        dissociated = new.update_coords(new_time=(None, new_time))
        assert dissociated.dim_map["new_time"] == ()


class TestSqueeze:
    """Tests for squeezing degenerate dimensions."""

    def test_bad_dim_raises(self, coord_manager_degenerate_time):
        """Ensure a bad dimension will raise CoordError."""
        with pytest.raises(CoordError, match="they don't exist"):
            coord_manager_degenerate_time.squeeze("is_money")

    def test_non_zero_length_dim_raises(self, coord_manager_degenerate_time):
        """Ensure a dim with length > 0 can't be squeezed."""
        with pytest.raises(CoordError, match="non-zero length"):
            coord_manager_degenerate_time.squeeze("distance")

    def test_squeeze_single_degenerate(self, coord_manager_degenerate_time):
        """Ensure a single degenerate dimension can be squeezed out."""
        cm = coord_manager_degenerate_time
        out = cm.squeeze("time")
        assert "time" not in out.dims

    def test_squeeze_no_dim(self, coord_manager_degenerate_time):
        """Ensure all degenerate dims are squeezed when no dim specified."""
        cm = coord_manager_degenerate_time
        out = cm.squeeze()
        assert "time" not in out.dims


class TestNonDimCoords:
    """Tests for adding non-dimensional coordinates."""

    def test_update_with_1d_coordinate(self, basic_coord_manager):
        """Ensure we can add coordinates."""
        lat = np.ones_like(basic_coord_manager.get_array("distance"))
        out = basic_coord_manager.update_coords(latitude=("distance", lat))
        assert out is not basic_coord_manager
        assert out.dims == basic_coord_manager.dims, "dims shouldn't change"
        assert np.all(out.get_array("latitude") == lat)
        assert out.dim_map["latitude"] == out.dim_map["distance"]
        # the dim map shouldn't shrink; only things get added.
        assert set(out.dim_map).issuperset(set(basic_coord_manager.dim_map))

    def test_init_with_1d_coordinate(self, basic_coord_manager):
        """Ensure initing with 1D non-dim coords works."""
        coords = dict(basic_coord_manager)
        lat = np.ones_like(basic_coord_manager.get_array("distance"))
        coords["latitude"] = ("distance", lat)
        out = get_coord_manager(coords, dims=basic_coord_manager.dims)
        assert out is not basic_coord_manager
        assert out.dims == basic_coord_manager.dims, "dims shouldn't change"
        assert np.all(out.get_array("latitude") == lat)
        assert out.dim_map["latitude"] == out.dim_map["distance"]
        # the dim map shouldn't shrink; only things get added.
        assert set(out.dim_map).issuperset(set(basic_coord_manager.dim_map))

    def test_update_2d_coord(self, basic_coord_manager):
        """Ensure updating can be done with 2D coordinate."""
        dist = basic_coord_manager.get_array("distance")
        time = basic_coord_manager.get_array("time")
        quality = np.ones((len(dist), len(time)))
        dims = ("distance", "time")
        new = basic_coord_manager.update_coords(qual=(dims, quality))
        assert new.dims == basic_coord_manager.dims
        assert new.dim_map["qual"] == dims
        assert "qual" in new


class TestDecimate:
    """Tests for evenly subsampling along a dimension."""

    def test_decimate_along_time(self, basic_coord_manager):
        """Simply ensure decimation reduces shape."""
        cm = basic_coord_manager
        ind = cm.dims.index("distance")
        out, _ = cm.decimate(distance=2)
        assert len(out.coord_map["distance"]) < len(cm.coord_map["distance"])
        assert (out.shape[ind] * 2) == cm.shape[ind]


class TestMergeCoordManagers:
    """Tests for merging coord managers together."""

    def _get_offset_coord_manager(self, cm, from_max=True, **kwargs):
        """Get a new coord manager offset by some amount along a dim."""
        name, value = next(iter(kwargs.items()))
        coord = cm.coord_map[name]
        start = coord.max() if from_max else coord.min()
        attr_name = f"{name}_min"
        new, _ = cm.update_from_attrs({attr_name: start + value})
        return new

    def test_merge_simple(self, basic_coord_manager):
        """Ensure we can merge simple, contiguous, coordinates together."""
        cm1 = basic_coord_manager
        time = cm1.coord_map["time"]
        cm2 = self._get_offset_coord_manager(cm1, time=time.step)
        out = merge_coord_managers([cm1, cm2], dim="time")
        new_time = out.coord_map["time"]
        assert isinstance(new_time, CoordRange)
        assert new_time.min() == time.min()
        assert new_time.max() == cm2.coord_map["time"].max()

    def test_merge_offset_close_no_snap(self, basic_coord_manager):
        """When the coordinate don't line up, it should produce monotonic Coord."""
        cm1 = basic_coord_manager
        dt = cm1.coord_map["time"].step
        # try a little more than dt
        cm2 = self._get_offset_coord_manager(cm1, time=dt * 1.1)
        out = merge_coord_managers([cm1, cm2], dim="time")
        assert isinstance(out.coord_map["time"], CoordMonotonicArray)
        # try a little less
        cm2 = self._get_offset_coord_manager(cm1, time=dt * 0.9)
        out = merge_coord_managers([cm1, cm2], dim="time")
        assert isinstance(out.coord_map["time"], CoordMonotonicArray)

    def test_merge_offset_overlap(self, basic_coord_manager):
        """Ensure coordinates that have overlap produce Coord Array."""
        cm1 = basic_coord_manager
        dt = cm1.coord_map["time"].step
        cm2 = self._get_offset_coord_manager(cm1, time=-dt * 1.1)
        out = merge_coord_managers([cm1, cm2], dim="time")
        assert isinstance(out.coord_map["time"], CoordArray)

    def test_merge_snap_but_not_needed(self, basic_coord_manager):
        """Specifying a snap tolerance even if coords line up should work."""
        cm1 = basic_coord_manager
        time = cm1.coord_map["time"]
        cm2 = self._get_offset_coord_manager(cm1, time=time.step)
        out = merge_coord_managers([cm1, cm2], dim="time", snap_tolerance=1.3)
        new_time = out.coord_map["time"]
        assert isinstance(new_time, CoordRange)
        assert new_time.min() == time.min()
        assert new_time.max() == cm2.coord_map["time"].max()

    @pytest.mark.parametrize("factor", (1.1, 0.9, 1.3, 1.01, 0))
    def test_merge_snap_when_needed(self, basic_coord_manager, factor):
        """Snap should be applied because when other cm is close expected."""
        cm1 = basic_coord_manager
        time = cm1.coord_map["time"]
        nt = time.step * factor
        cm2 = self._get_offset_coord_manager(cm1, time=nt)
        out = merge_coord_managers([cm1, cm2], dim="time", snap_tolerance=1.3)
        new_time = out.coord_map["time"]
        assert isinstance(new_time, CoordRange)
        assert new_time.min() == time.min()
        new_dim_len = out.shape[cm1.dims.index("time")]
        expected_end = time.min() + (new_dim_len - 1) * time.step
        assert new_time.max() == expected_end

    @pytest.mark.parametrize("factor", (10, -10, 6, -6))
    def test_merge_raise_snap_too_big(self, basic_coord_manager, factor):
        """When snap is too big, an error should be raised."""
        cm1 = basic_coord_manager
        time = cm1.coord_map["time"]
        nt = time.step * factor
        cm2 = self._get_offset_coord_manager(cm1, time=nt)
        with pytest.raises(CoordMergeError, match="Snap tolerance"):
            merge_coord_managers([cm1, cm2], dim="time", snap_tolerance=1.3)

    def test_different_dims_raises(self, basic_coord_manager):
        """When dimensions differ merge should raise."""
        cm1 = basic_coord_manager
        cm2 = cm1.rename_coord(distance="dist")
        with pytest.raises(CoordMergeError, match="same dimensions"):
            merge_coord_managers([cm1, cm2], "time")
        with pytest.raises(CoordMergeError, match="same dimensions"):
            merge_coord_managers([cm1, cm2], "distance")

    def test_different_units_raises(self, basic_coord_manager):
        """When dimensions differ merge should raise."""
        cm1 = basic_coord_manager
        cm2 = cm1.set_units(distance="furlong")
        with pytest.raises(CoordMergeError, match="share the same units"):
            merge_coord_managers([cm1, cm2], "distance")

    def test_unequal_non_merge_coords(self, basic_coord_manager):
        """When coords that won't be merged arent equal merge should fail."""
        cm1 = basic_coord_manager
        dist = cm1.coord_map["distance"]
        new_dist = dist.update_limits(min=dist.max())
        cm2 = cm1.update_coords(distance=new_dist)
        with pytest.raises(CoordMergeError, match="Non merging coordinates"):
            merge_coord_managers([cm1, cm2], "time")

    def test_unshared_coord_dropped(self, basic_coord_manager):
        """
        When one coord manager has coords and the other doesn't they should
        be dropped.
        """
        cm1 = basic_coord_manager
        cm2 = cm1.update_coords(time2=("time", cm1.get_array("time")))
        out_no_range = merge_coord_managers([cm1, cm2], "time")
        assert "time2" not in out_no_range.coord_map
        out_with_range = merge_coord_managers([cm1, cm2], "time")
        assert "time2" not in out_with_range.coord_map

    def test_slightly_different_dt(self, coord_dt_small_diff):
        """
        Ensure coord managers with slightly different dt can still merge
        but produce uneven sampled dimension.
        """
        cm = coord_dt_small_diff
        coord = cm.coord_map["time"]
        assert coord.sorted


class TestSort:
    """Tests for sorting coord managers."""

    def test_sort_unsorted_dim_coord(self, coord_manager_wacky_dims):
        """Simple test for single dim sort."""
        cm = coord_manager_wacky_dims
        sorted_cm, _ = cm.sort("distance")
        assert sorted_cm.coord_map["distance"].sorted

    def test_sort_sorted_dim_coord(self, coord_manager_wacky_dims):
        """Ensure sorting a sorted dim coord does nothing."""
        cm = coord_manager_wacky_dims
        sorted_cm, _ = cm.sort("time")
        assert sorted_cm.coord_map["time"].sorted

    def test_r_sort_unsorted_dim_coord(self, coord_manager_wacky_dims):
        """Simple test for single dim sort."""
        cm = coord_manager_wacky_dims
        sorted_cm, _ = cm.sort("distance", reverse=True)
        assert sorted_cm.coord_map["distance"].reverse_sorted

    def test_r_sort_sorted_dim_coord(self, coord_manager_wacky_dims):
        """Ensure sorting a sorted dim coord does nothing."""
        cm = coord_manager_wacky_dims
        sorted_cm, _ = cm.sort("time", reverse=True)
        assert sorted_cm.coord_map["time"].reverse_sorted

    def test_simultaneous_sort(self, coord_manager_wacky_dims):
        """Ensure all dimensions can be sorted at once."""
        sorted_cm, _ = coord_manager_wacky_dims.sort()
        for _, coord in sorted_cm.coord_map.items():
            assert coord.sorted

    def test_simultaneous_r_sort(self, coord_manager_wacky_dims):
        """Ensure all dimensions can be reverse sorted at once."""
        sorted_cm, _ = coord_manager_wacky_dims.sort(reverse=True)
        for _, coord in sorted_cm.coord_map.items():
            assert coord.reverse_sorted

    def test_sort_sorted_cm(self, basic_coord_manager):
        """Sorting a coord manager that is already sorted should do nothing."""
        cm = basic_coord_manager
        cm_sorted, _ = cm.sort()
        assert cm == cm_sorted

    def test_sort_2d_coord_raises(self, coord_manager_multidim):
        """Sorting on 2D coord is ambiguous, it should raise."""
        cm = coord_manager_multidim
        with pytest.raises(CoordSortError, match="more than one dimension"):
            cm.sort("quality")

    def test_sort_two_coords_same_dim_raises(self, coord_manager_multidim):
        """Trying to sort on two coords which share a dim should raise."""
        # need to make distance un sorted.
        cm, _ = coord_manager_multidim.sort("distance", reverse=True)
        with pytest.raises(CoordSortError, match="they share a dimension"):
            cm.sort("latitude", "distance")
        # this should also raise since sorting latitude could unsort distance
        with pytest.raises(CoordSortError, match="they share a dimension"):
            coord_manager_multidim.sort("distance", "latitude")

    def test_related_coords(self, coord_manager_multidim):
        """Sorting on one coordinate should also sort others that share a dim."""
        cm = coord_manager_multidim
        cm_sorted, _ = cm.sort("latitude")
        assert cm_sorted.coord_map["latitude"].sorted
        # distance should now be unsorted.
        assert not cm_sorted.coord_map["distance"].sorted


class TestSnap:
    """Tests for snapping coordinates."""

    def test_generic_snap(self, coord_manager):
        """All coord managers should support snap."""
        cm, _ = coord_manager.snap()
        for dim in cm.dims:
            coord = cm.coord_map[dim]
            if coord.degenerate or len(coord) < 2:
                continue
            assert coord.sorted
            assert not coord.reverse_sorted
            assert coord.evenly_sampled

    def test_generic_snap_reverse(self, coord_manager):
        """All coord managers should support snap in reverse direction."""
        cm, _ = coord_manager.snap(reverse=True)
        for dim in cm.dims:
            coord = cm.coord_map[dim]
            if coord.degenerate or len(coord) < 2:
                continue
            assert not coord.sorted
            assert coord.reverse_sorted
            assert coord.evenly_sampled

    def test_snap_dims(self, coord_manager_wacky_dims):
        """Happy path for snapping dimensions."""
        cm = coord_manager_wacky_dims
        data = np.ones(cm.shape)
        out, out_data = cm.snap(array=data)
        for coord_name in out.coord_map:
            coord1 = cm.coord_map[coord_name]
            coord2 = out.coord_map[coord_name]
            assert coord1.dtype == coord2.dtype
            assert len(coord1) == len(coord2)
        assert out_data.shape == data.shape

    def test_snap_expected_dt(
        self, coord_dt_small_diff, memory_spool_small_dt_differences
    ):
        """Ensure snapping creates expected dt from merged df."""
        spool = memory_spool_small_dt_differences
        expected_dt = get_middle_value([x.attrs.time_step for x in spool])
        snapped = coord_dt_small_diff.snap()[0]
        assert snapped.coord_map["time"].step == expected_dt


class TestConvertUnits:
    """Tests for converting coordinate units."""

    def test_convert_changes_labels(self, basic_coord_manager):
        """Basic tests for converting units."""
        cm = basic_coord_manager.convert_units(time="s", distance="furlong")
        time, dist = cm.coord_map["time"], cm.coord_map["distance"]
        assert time.units == get_quantity("s")
        assert dist.units == get_quantity("furlong")

    def test_convert_changes_values(self, coord_manager_with_units):
        """Ensure values are scaled accordingly."""
        conv = get_quantity("m").to("ft").magnitude
        cm = coord_manager_with_units.convert_units(distance="ft")
        dist1 = coord_manager_with_units.coord_map["distance"]
        dist2 = cm.coord_map["distance"]
        assert np.isclose(dist1.step, dist2.step / conv)
        assert np.isclose(dist1.start, dist2.start / conv)
        assert np.isclose(dist1.stop, dist2.stop / conv)

    def test_convert_time(self, coord_manager_with_units):
        """When time is already set and a datetime, units should just change."""
        # TODO, reconsider this; i am not sure its right.
        new = coord_manager_with_units.convert_units(time="ms")
        time1 = coord_manager_with_units.coord_map["time"]
        time2 = new.coord_map["time"]
        assert time1.dtype == time2.dtype
        assert np.all(np.equal(time1.values, time2.values))


class TestDisassociate:
    """Ensure coordinates can be disassociated from coordinate manager."""

    def test_basic_disassociate(self, coord_manager_multidim):
        """Ensure coordinates can be dissociated."""
        cm = coord_manager_multidim.disassociate_coord("time")
        # now time should still exist but have no dim, quality should drop
        assert "time" in cm.coord_map
        assert "quality" not in cm.coord_map
        time_dim = cm.dim_map["time"]
        assert time_dim == ()
        assert "time" not in cm.dims
        # but if both time and quality are dissociated nothing should be dropped.
        cm = coord_manager_multidim.disassociate_coord("time", "quality")
        assert {"time", "quality"}.issubset(set(cm.coord_map))
        assert "time" not in cm.dims

    def test_disassociate_non_dim_coord(self, coord_manager_multidim):
        """Ensure non-dim coords can be dissociated with no change to dims."""
        cm = coord_manager_multidim.disassociate_coord("quality")
        assert cm.dim_map["quality"] == ()
        assert cm.dims == coord_manager_multidim.dims


class TestSetDims:
    """Tests for setting dims to non-dimensional coordinates."""

    def test_swap_coords(self, coord_manager_multidim):
        """Simple coord swap."""
        cm = coord_manager_multidim
        out1 = cm.set_dims(distance="latitude")
        assert "latitude" in out1.dims
        assert "latitude" not in cm.dims, "original dim should be unchanged"
        # distance should now be associated with latitude.
        assert out1.dim_map["distance"][0] == "latitude"

    def test_bad_dimension(self, coord_manager_multidim):
        """Ensure if a bad dimension is provided it raises."""
        cm = coord_manager_multidim
        match = "is not a dimension or"
        with pytest.raises(CoordError, match=match):
            cm.set_dims(bob="latitude")
        with pytest.raises(CoordError, match=match):
            cm.set_dims(distance="bob")

    def test_wrong_shape(self, coord_manager_multidim):
        """Ensure if a coord with wrong shape is chosen an error is raised."""
        cm = coord_manager_multidim
        match = "does not match the shape of"
        with pytest.raises(CoordError, match=match):
            cm.set_dims(distance="quality")



"""Tests for IO utilities."""
from __future__ import annotations

from io import BufferedReader, BufferedWriter
from pathlib import Path

import pytest
from tables import File

import dascore as dc
from dascore.exceptions import PatchConversionError
from dascore.utils.hdf5 import HDF5Reader, HDF5Writer
from dascore.utils.io import (
    BinaryReader,
    BinaryWriter,
    IOResourceManager,
    get_handle_from_resource,
)


class _BadType:
    """A dummy type for testing."""


def _dummy_func(arg: str, arg2: _BadType) -> int:
    """A dummy function."""


class TestGetHandleFromResource:
    """Tests for getting the file handle from specific resources."""

    def test_bad_type(self):
        """
        In order to not break anything, unsupported types should just
        return the original argument.
        """
        out = get_handle_from_resource("here", _BadType)
        assert out == "here"

    def test_path_to_buffered_reader(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_read_buffer.txt"
        path.touch()
        handle = get_handle_from_resource(path, BinaryReader)
        assert isinstance(handle, BufferedReader)
        handle.close()

    def test_path_to_buffered_writer(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_buffered_writer.txt"
        handle = get_handle_from_resource(path, BinaryWriter)
        assert isinstance(handle, BufferedWriter)
        handle.close()

    def test_path_to_hdf5_reader(self, generic_hdf5):
        """Ensure we get a reader from tmp path reader."""
        handle = get_handle_from_resource(generic_hdf5, HDF5Reader)
        assert isinstance(handle, File)
        handle.close()

    def test_path_to_hdf5_writer(self, tmp_path):
        """Ensure we get a reader from tmp path reader."""
        path = tmp_path / "test_hdf_writer.h5"
        handle = get_handle_from_resource(path, HDF5Writer)
        assert isinstance(handle, File)

    def test_get_path(self, tmp_path):
        """Ensure we can get a path."""
        path = get_handle_from_resource(tmp_path, Path)
        assert isinstance(path, Path)

    def test_get_str(self, tmp_path):
        """Ensure we can get a string."""
        my_str = get_handle_from_resource(tmp_path, str)
        assert isinstance(my_str, str)

    def test_already_file_handle(self, tmp_path):
        """Ensure an input that is already the requested type works."""
        path = tmp_path / "pass_back.txt"
        with open(path, "wb") as fi:
            out = get_handle_from_resource(fi, BinaryWriter)
            assert out is fi

    def test_not_implemented(self):
        """Tests for raising not implemented errors for types not supported."""
        bad_instance = _BadType()
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, BinaryReader)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, BinaryWriter)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, HDF5Writer)
        with pytest.raises(NotImplementedError):
            get_handle_from_resource(bad_instance, HDF5Reader)


class TestIOResourceManager:
    """Tests for the IO resource manager."""

    def test_basic_context_manager(self, tmp_path):
        """Ensure it works as a context manager."""
        write_path = tmp_path / "io_writer"

        with IOResourceManager(write_path) as man:
            my_str = man.get_resource(_dummy_func)
            assert isinstance(my_str, str)
            path = man.get_resource(Path)
            assert isinstance(path, Path)
            hf = man.get_resource(HDF5Writer)
            fi = man.get_resource(BinaryWriter)
            # Why didn't pytables implement the stream like pythons?
            assert hf.isopen
            assert not fi.closed
        # after the context manager exists everything should be closed.
        assert not hf.isopen
        assert fi.closed

    def test_nested_context(self, tmp_path):
        """Ensure nested context works as well."""
        write_path = tmp_path / "io_writer"
        with IOResourceManager(write_path) as man:
            fi1 = man.get_resource(BinaryWriter)
            with IOResourceManager(man):
                fi2 = man.get_resource(BinaryWriter)
                # nested IOManager should just return value from previous
                assert fi1 is fi2
            # on first exist the resource should remain open
            assert not fi2.closed
        # then closed.
        assert fi2.closed

    def test_closed_after_exception(self, tmp_path):
        """Ensure the file resources are closed after an exception."""
        path = tmp_path / "closed_resource_test.txt"
        path.touch()
        try:
            with IOResourceManager(path) as man:
                fi = man.get_resource(BinaryReader)
                raise ValueError("Waaagh!")
        except ValueError:
            assert fi.closed


class TestXarray:
    """Tests for xarray conversions."""

    @pytest.fixture
    def data_array_from_patch(self, random_patch):
        """Get a data array from a patch."""
        pytest.importorskip("xarray")
        return random_patch.io.to_xarray()

    def test_convert_to_xarray(self, data_array_from_patch):
        """Tests for converting to xarray object."""
        import xarray as xr

        assert isinstance(data_array_from_patch, xr.DataArray)

    def test_convert_from_xarray(self, data_array_from_patch):
        """Ensure xarray data arrays can be converted back."""
        out = dc.utils.io.xarray_to_patch(data_array_from_patch)
        assert isinstance(out, dc.Patch)

    def test_round_trip(self, random_patch, data_array_from_patch):
        """Converting to xarray should be lossless."""
        out = dc.utils.io.xarray_to_patch(data_array_from_patch)
        assert out == random_patch


class TestObsPy:
    """Tests for converting patches to/from ObsPy streams."""

    @pytest.fixture
    def short_patch(self, random_patch):
        """Just shorten the patch distance dim to speed up these tests."""
        return random_patch.select(distance=(0, 10), samples=True)

    @pytest.fixture
    def stream_from_patch(self, short_patch):
        """Get a stream from a patch."""
        pytest.importorskip("obspy")
        st = short_patch.io.to_obspy()
        return st

    def test_convert_to_obspy(self, stream_from_patch):
        """Ensure a patch can be converted to a stream."""
        import obspy

        assert isinstance(stream_from_patch, obspy.Stream)

    def test_obspy_to_patch(self, stream_from_patch):
        """Ensure we can convert back to patch from stream."""
        out = dc.io.obspy_to_patch(stream_from_patch)
        assert isinstance(out, dc.Patch)

    def test_patch_no_time_raises(self, random_patch):
        """Ensure a patch without time dimension raises."""
        pytest.importorskip("obspy")
        patch = random_patch.rename_coords(time="not_time")
        with pytest.raises(PatchConversionError):
            patch.io.to_obspy()

    def test_bad_stream_raises(self):
        """Ensure a stream without even length or require param raises."""
        obspy = pytest.importorskip("obspy")
        st = obspy.read()
        # since st doesn't have a value of "distance" in each of its traces
        # attrs dict this should raise.
        with pytest.raises(PatchConversionError):
            dc.io.obspy_to_patch(st)

    def test_empty_stream(self):
        """An empty Stream should return an empty Patch."""
        obspy = pytest.importorskip("obspy")
        st = obspy.Stream([])
        patch = dc.io.obspy_to_patch(st)
        assert not patch.dims



"""
A module for applying transformation to Patches.

Transforms are defined as
"""
from __future__ import annotations

from .differentiate import differentiate
from .fft import rfft
from .fourier import dft, idft
from .integrate import integrate
from .spectro import spectrogram
from .strain import velocity_to_strain_rate
from .dispersion import dispersion_phase_shift



"""Tests for performing aggregations."""
from __future__ import annotations

import numpy as np
import pytest

from dascore.proc.aggregate import _AGG_FUNCS, aggregate


class TestBasicAggregations:
    """Sanity checks for basic aggregations."""

    @pytest.fixture(params=list(_AGG_FUNCS))
    def distance_aggregated_patch(self, request, random_patch):
        """Apply all supported aggregations along distance axis."""
        agg = request.param
        return aggregate(random_patch, dim="distance", method=agg)

    def test_dimension_collapsed(self, distance_aggregated_patch):
        """Ensure the aggregate dimension was collapsed to len 1."""
        axis = distance_aggregated_patch.dims.index("distance")
        axis_len = distance_aggregated_patch.data.shape[axis]
        assert axis_len == 1
        dist = distance_aggregated_patch.coords.get_array("distance")
        assert len(dist) == axis_len

    def test_time_aggregate(self, random_patch):
        """Tests for aggregating on time axis."""
        out = random_patch.aggregate(dim="time")
        assert out.attrs["time_max"] == out.attrs["time_min"]
        assert 1 in out.data.shape

    def test_first(self, random_patch):
        """Ensure aggregations can occur."""
        out = random_patch.aggregate(dim="distance", method="first")
        axis = random_patch.dims.index("distance")
        assert out.data.shape[axis] == 1
        assert np.allclose(random_patch.data[0, :], out.data[0, :])

    def test_last(self, random_patch):
        """Ensure aggregations can occur."""
        out = random_patch.aggregate(dim="distance", method="last")
        axis = random_patch.dims.index("distance")
        assert out.data.shape[axis] == 1
        assert np.allclose(random_patch.data[-1, :], out.data[0, :])



"""
dascore Filtering module.

Much of this code was inspired by ObsPy's filtering module created by:
Tobias Megies, Moritz Beyreuther, Yannik Behr
"""
from __future__ import annotations

import pandas as pd
from scipy import ndimage
from scipy.ndimage import median_filter as nd_median_filter
from scipy.signal import iirfilter, sosfilt, sosfiltfilt, zpk2sos

import dascore
from dascore.constants import PatchType, samples_arg_description
from dascore.exceptions import FilterValueError
from dascore.units import get_filter_units
from dascore.utils.docs import compose_docstring
from dascore.utils.misc import check_filter_kwargs, check_filter_range
from dascore.utils.patch import (
    get_dim_sampling_rate,
    get_multiple_dim_value_from_kwargs,
    patch_function,
)


def _check_sobel_args(dim, mode, cval):
    """Check Sobel filter kwargs and return."""
    mode_options = {
        "reflect",
        "constant",
        "nearest",
        "mirror",
        "wrap",
        "grid-constant",
        "grid-mirror",
        "grid-wrap",
    }
    if not isinstance(dim, str):
        msg = "dim parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(mode, str):
        msg = "mode parameter should be a string."
        raise FilterValueError(msg)
    if not isinstance(cval, float | int):
        msg = "cval parameter should be a float or an int."
        raise FilterValueError(msg)
    if mode not in mode_options:
        msg = f"The valid values for modes are {mode_options}."
        raise FilterValueError(msg)

    return dim, mode, cval


def _get_sos(sr, filt_min, filt_max, corners):
    """Get second order sections from sampling rate and filter bounds."""
    nyquist = 0.5 * sr
    low = None if pd.isnull(filt_min) else filt_min / nyquist
    high = None if pd.isnull(filt_max) else filt_max / nyquist
    check_filter_range(nyquist, low, high, filt_min, filt_max)

    if (low is not None) and (high is not None):  # apply bandpass
        z, p, k = iirfilter(
            corners, [low, high], btype="band", ftype="butter", output="zpk"
        )
    elif low is not None:
        z, p, k = iirfilter(
            corners, low, btype="highpass", ftype="butter", output="zpk"
        )
    else:
        assert high is not None
        z, p, k = iirfilter(
            corners, high, btype="lowpass", ftype="butter", output="zpk"
        )
    return zpk2sos(z, p, k)


@patch_function()
def pass_filter(patch: PatchType, corners=4, zerophase=True, **kwargs) -> PatchType:
    """
    Apply a Butterworth pass filter (bandpass, highpass, or lowpass).

    Parameters
    ----------
    corners
        The number of corners for the filter.
    zerophase
        If True, apply the filter twice.
    **kwargs
        Used to specify the dimension and frequency, wavelength, or equivalent
        limits.

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # Apply bandpass filter along time axis from 1 to 100 Hz
    >>> bandpassed = pa.pass_filter(time=(1, 100))

    >>>  # Apply lowpass filter along distance axis for wavelengths less than 100m
    >>> lowpassed = pa.pass_filter(distance=(None, 1/100))
    >>> # Note that None and ... both indicate open intervals
    >>> assert pa.pass_filter(time=(None, 90)) == pa.pass_filter(time=(..., 90))

    >>> # Optionally, units can be specified for a more expressive API.
    >>> from dascore.units import m, ft, s, Hz
    >>> # Filter from 1 Hz to 10 Hz in time dimension
    >>> lp_units = pa.pass_filter(time=(1 * Hz, 10 * Hz))
    >>> # Filter wavelengths 50m to 100m
    >>> bp_m = pa.pass_filter(distance=(50 * m, 100 * m))
    >>> # filter wavelengths less than 200 ft
    >>> lp_ft = pa.pass_filter(distance=(200 * ft, ...))
    """
    dim, (arg1, arg2) = check_filter_kwargs(kwargs)
    axis = patch.dims.index(dim)
    coord_units = patch.coords.coord_map[dim].units
    filt_min, filt_max = get_filter_units(arg1, arg2, to_unit=coord_units, dim=dim)
    sr = get_dim_sampling_rate(patch, dim)
    # get nyquist and low/high in terms of nyquist
    sos = _get_sos(sr, filt_min, filt_max, corners)
    if zerophase:
        out = sosfiltfilt(sos, patch.data, axis=axis)
    else:
        out = sosfilt(sos, patch.data, axis=axis)
    return dascore.Patch(
        data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims
    )


@patch_function()
def sobel_filter(patch: PatchType, dim: str, mode="reflect", cval=0.0) -> PatchType:
    """
    Apply a Sobel filter.

    Parameters
    ----------
    dim
        The dimension along which to apply
    mode
        Determines how the input array is extended when the filter
        overlaps with a border.
    cval
        Fill value when mode="constant".

    Examples
    --------
    >>> import dascore
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply Sobel filter using the default parameter values.
    >>> sobel_default = pa.sobel_filter(dim='time', mode='reflect', cval=0.0)

    >>>  # 2. Apply Sobel filter with arbitrary parameter values.
    >>> sobel_arbitrary = pa.sobel_filter(dim='time', mode='constant', cval=1)

    >>> # 3. Apply Sobel filter along both axes
    >>> sobel_time_space = pa.sobel_filter('time').sobel_filter('distance')
    """
    dim, mode, cval = _check_sobel_args(dim, mode, cval)
    axis = patch.dims.index(dim)
    out = ndimage.sobel(patch.data, axis=axis, mode=mode, cval=cval)
    return dascore.Patch(
        data=out, coords=patch.coords, attrs=patch.attrs, dims=patch.dims
    )


#
# @patch_function()
# def stop_filter(patch: PatchType, corners=4, zerophase=True, **kwargs) -> PatchType:
#     """
#     Apply a Butterworth band stop filter or (highpass, or lowpass).
#
#     Parameters
#     ----------
#     corners
#         The number of corners for the filter.
#     zerophase
#         If True, apply the filter twice.
#     **kwargs
#         Used to specify the dimension and frequency, wavenumber, or equivalent
#         limits.
#
#
#     """
#     # get nyquist and low/high in terms of nyquist
#     if zerophase:


def _create_size_and_axes(patch, kwargs, samples):
    """Return"""
    dimfo = get_multiple_dim_value_from_kwargs(patch, kwargs)
    axes = [x["axis"] for x in dimfo.values()]
    size = [1] * len(patch.dims)
    for dim, info in dimfo.items():
        axis = info["axis"]
        coord = patch.get_coord(dim)
        window = coord.get_sample_count(info["value"], samples=samples)
        size[axis] = window
    return tuple(size), tuple(axes)


@patch_function()
@compose_docstring(sample_explination=samples_arg_description)
def median_filter(
    patch: PatchType, samples=False, mode="reflect", cval=0.0, **kwargs
) -> PatchType:
    """
    Apply 2-D median filter.

    Parameters
    ----------
    patch
        The patch to filter
    samples
        {sample_explination}
    mode
        The mode for handling edges.
    cval
        The constant value for when mode == constant.
    **kwargs
        Used to specify the shape of the median filter in each dimension.
        See examples for more info.

    Examples
    --------
    >>> import dascore
    >>> from dascore.units import m, s
    >>> pa = dascore.get_example_patch()

    >>>  # 1. Apply median filter only over time distance with 0.10 sec window
    >>> filtered_pa_1 = pa.median_filter(time=0.1)

    >>>  # 2. Apply median filter over both time and distance
    >>>  # using a 0.1 second time window and 2 m distance window
    >>> filtered_pa_2 = pa.median_filter(time=0.1 * s, distance=2 * m)

    >>>  # 3. Apply median filter with 3 time samples and 4 distance samples
    >>> filtered_pa = pa.median_filter(
    ...     time=3, distance=4, samples=True,
    ... )

    Notes
    -----
    See scipy.ndimage.median_filter for more info on implementation
    and arguments.

    Values specified with kwargs should be small, for example < 10 samples
    otherwise this can take a long time and use lots of memory.
    """
    size, _ = _create_size_and_axes(patch, kwargs, samples)
    new_data = nd_median_filter(patch.data, size=size, mode=mode, cval=cval)
    return patch.new(data=new_data)



"""
Modules for reading and writing fiber data.
"""
from __future__ import annotations

import dascore.utils.pd
from dascore.io.core import FiberIO, read, scan, scan_to_df, write, PatchFileSummary
from dascore.utils.io import BinaryReader, BinaryWriter
from dascore.utils.hdf5 import (
    HDF5Writer,
    HDF5Reader,
    PyTablesWriter,
    PyTablesReader,
    H5Reader,
    H5Writer,
)
from dascore.utils.misc import MethodNameSpace
from dascore.utils.pd import dataframe_to_patch, patch_to_dataframe
from dascore.utils.io import (
    xarray_to_patch,
    patch_to_xarray,
    patch_to_obspy,
    obspy_to_patch,
)


class PatchIO(MethodNameSpace):
    write = write
    to_dataframe = patch_to_dataframe
    to_xarray = patch_to_xarray
    to_obspy = patch_to_obspy



"""Tests for example fetching."""
from __future__ import annotations

import numpy as np
import pytest

import dascore as dc
from dascore.exceptions import UnknownExample
from dascore.utils.time import to_float


class TestGetExamplePatch:
    """Test suite for `get_example_patch`."""

    def test_default(self):
        """Ensure calling get_example_patch with no args returns patch."""
        patch = dc.get_example_patch()
        assert isinstance(patch, dc.Patch)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExample, match="No example patch"):
            dc.get_example_patch("NotAnExampleRight????")

    def test_example_1(self):
        """Ensure example 1 returns a Patch."""
        out = dc.get_example_patch("example_event_1")
        assert isinstance(out, dc.Patch)

    def test_example_2(self):
        """Ensure example 2 returns a Patch."""
        out = dc.get_example_patch("example_event_2")
        assert isinstance(out, dc.Patch)

    def test_sin_wav(self):
        """Ensure the sin wave example can be loaded. See issee 229."""
        out = dc.get_example_patch("sin_wav")
        assert isinstance(out, dc.Patch)


class TestGetExampleSpool:
    """Test suite for `get_example_spool`."""

    def test_default(self):
        """Ensure calling get_example_spool with no args returns a Spool."""
        patch = dc.get_example_spool()
        assert isinstance(patch, dc.BaseSpool)

    def test_raises_on_bad_key(self):
        """Ensure a bad key raises expected error."""
        with pytest.raises(UnknownExample, match="No example spool"):
            dc.get_example_spool("NotAnExampleRight????")


class TestRickerMoveout:
    """Tests for Ricker moveout patch."""

    def test_moveout(self):
        """Ensure peaks of ricker wavelet line up with expected moveout."""
        velocity = 100
        patch = dc.get_example_patch("ricker_moveout", velocity=velocity)
        argmaxes = np.argmax(patch.data, axis=0)
        peak_times = patch.get_coord("time").values[argmaxes]
        moveout = to_float(peak_times - np.min(peak_times))
        distances = patch.get_coord("distance").values
        expected_moveout = distances / velocity
        assert np.allclose(moveout, expected_moveout)



"""Processing for applying roller operations."""
from __future__ import annotations

from typing import Any, Literal

import numpy as np
import pandas as pd
from pydantic import Field

import dascore as dc
from dascore.constants import samples_arg_description
from dascore.exceptions import ParameterError
from dascore.utils.docs import compose_docstring
from dascore.utils.models import DascoreBaseModel
from dascore.utils.patch import get_dim_value_from_kwargs


class _PatchRollerInfo(DascoreBaseModel):
    """
    A dataclass for storing info on rolling operation.

    Should be subclassed to implement rolling methods.
    """

    patch: Any  # cant set to patch due to circular import
    window: int
    step: int
    dim: str
    axis: int
    center: bool
    roll_hist: str = ""
    func_kwargs: dict = Field(default_factory=dict)

    def get_coords(self):
        """
        Get the new coordinates for "rolled" patch.

        Accounts for centered or non-centered coordinates. If the window
        length is even, the first half value is used.
        """
        coord = self.patch.get_coord(self.dim)
        if self.step > 1:
            coord = coord[:: self.step]
        return self.patch.coords.update_coords(**{self.dim: coord})

    def _get_attrs_with_apply_history(self, func_or_str):
        """Get new attrs that has history from apply attached."""
        new_history = list(self.patch.attrs.history)
        if callable(func_or_str):
            func_name = getattr(func_or_str, "__name__", "")
            hist_str = f"{self.roll_hist}.apply({func_name})"
        else:
            hist_str = f"{self.roll_hist}.{func_or_str}()"
        new_history.append(hist_str)
        attrs = self.patch.attrs.update(history=new_history, coords={})
        return attrs


class _NumpyPatchRoller(_PatchRollerInfo):
    """A class to apply roller operations to patches."""

    def get_start_index(self):
        """
        Get the start index to account for non-zero step size.

        This only applies for numpy engine.
        """
        wsize = self.window - 1
        out = np.ceil(wsize / self.step) * self.step - wsize
        return int(out)

    def _pad_roll_array(self, data):
        """Pad."""
        num_nans = 1 + (self.window - 2) // self.step
        pad_width = [(0, 0)] * len(data.shape)
        pad_width[self.axis] = (num_nans, 0)
        padded = np.pad(data, pad_width, constant_values=np.NaN)
        if self.step == 1:
            assert padded.shape == self.patch.data.shape
        if self.center:
            # roll array along axis to center
            padded = np.roll(padded, -self.window // 2, axis=self.axis)
        return padded

    def apply(self, function):
        """
        Apply a function over the specified moving window.

        Parameters
        ----------
        function
            The function which is applied. Must accept an axis argument.
        """
        # TODO look at replacing this with a call to `as_strided` that
        # accounts for strides.
        slide_view = np.lib.stride_tricks.sliding_window_view(
            self.patch.data,
            self.window,
            self.axis,
        )
        # get slice to account for step (stride)
        step_slice = [slice(None, None)] * len(self.patch.data.shape)
        step_slice.append(slice(None, None))
        # this accounts for NaNs that pad the start of the array.
        start = self.get_start_index()
        # start = (self.window - 1) % self.step
        step_slice[self.axis] = slice(start, None, self.step)
        # apply function, then pad with zeros and roll
        kwargs = self.func_kwargs
        trimmed_slide_view = slide_view[tuple(step_slice)]
        raw = function(trimmed_slide_view, axis=-1, **kwargs).astype(np.float_)
        out = self._pad_roll_array(raw)
        new_coords = self.get_coords()
        attrs = self._get_attrs_with_apply_history(function)
        return self.patch.new(data=out, coords=new_coords, attrs=attrs)

    def mean(self):
        """Apply mean to moving window."""
        return self.apply(np.mean)

    def median(self):
        """Apply median to moving window."""
        return self.apply(np.median)

    def min(self):
        """Apply min to moving window."""
        return self.apply(np.min)

    def max(self):
        """Apply max to moving window."""
        return self.apply(np.max)

    def std(self):
        """Apply standard deviation to moving window."""
        return self.apply(np.std)

    def sum(self):
        """Apply sum to moving window."""
        return self.apply(np.sum)


class _PandasPatchRoller(_PatchRollerInfo):
    """A class to apply pandas rolling operations."""

    def _get_df(self) -> pd.DataFrame:
        """Get the dataframe from patch data."""
        if len(self.patch.dims) > 2:
            msg = "Cannot use Pandas engine on patches with more than 2 dims."
            raise ParameterError(msg)
        df = pd.DataFrame(self.patch.data)
        return df

    def _get_rolling(self):
        """Get rolling."""
        df = self._get_df()
        roll = df.rolling(
            window=self.window,
            step=self.step,
            axis=self.axis,
            center=self.center,
        )
        return roll

    def _repack_patch(self, df, attrs=None):
        """Repack patch into dataframe."""
        data = df.values
        # get rid of extra dims if original data doesn't have them.
        if len(data.shape) != len(self.patch.data.shape):
            data = np.squeeze(data)
        coords = self.get_coords()
        return self.patch.new(data=data, coords=coords, attrs=attrs)

    def _call_rolling_func(self, name, *args, **kwargs):
        """Helper function for calling a rolling function."""
        rolling = self._get_rolling()
        df = getattr(rolling, name)(*args, **kwargs)
        attrs = self._get_attrs_with_apply_history(name)
        return self._repack_patch(df, attrs=attrs)

    def apply(self, func):
        df = self._get_rolling().apply(func, **self.func_kwargs)
        attrs = self._get_attrs_with_apply_history(func)
        return self._repack_patch(df, attrs=attrs)

    def mean(self):
        """Apply mean."""
        return self._call_rolling_func(name="mean")

    def median(self):
        """Apply median to moving window."""
        return self._call_rolling_func(name="median")

    def min(self):
        """Apply min to moving window."""
        return self._call_rolling_func(name="min")

    def max(self):
        """Apply max to moving window."""
        return self._call_rolling_func(name="max")

    def std(self):
        """Apply standard deviation to moving window."""
        return self._call_rolling_func(name="std")

    def sum(self):
        """Apply sum to moving window."""
        return self._call_rolling_func(name="sum")


@compose_docstring(sample_explination=samples_arg_description)
def rolling(
    patch: dc.Patch,
    step=None,
    center=False,
    engine: Literal["numpy", "pandas", None] = None,
    samples=False,
    **kwargs,
) -> _NumpyPatchRoller:
    """
    Apply a rolling function along a specified dimension.

    Parameters
    ----------
    step
        The window is evaluated at every step result, equivalent to slicing
        at every step. If the step argument is not None, the result will
        have a different shape than the input.
    center
        If False, set the window labels as the right edge of the window index.
        If True, set the window labels as the center of the window index.
    engine
        Determines how the rolling operations are applied. If None, try to
        determine which will be fastest for a given step. Options are:
            "numpy" - which uses np.lib.stride_tricks.sliding_window_view.
            "pandas" - which uses pandas.rolling.
        If step < 10 samples, pandas is faster for all operations other than apply.
        If step > 10 samples, or `apply` is the desired rolling operation, numpy
        is probably better.
    samples
        {sample_explination}
    **kwargs
        Used to pass dimension and window size.
        For example `time=10` represents window size of
        10*(default unit) along the time axis.

    Examples
    --------
    >>> # Simple example for rolling mean function
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # apply rolling over 1 second with 0.5 step
    >>> mean_patch = patch.rolling(time=1, step=0.5).mean()
    >>> # drop nan at the start of the time axis.
    >>> out = mean_patch.dropna("time")

    Notes
    -----
    This class behaves like pandas.rolling
    (https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.rolling.html)
    which has some important implications.

    First, when step is not defined or set to 1, the output patch will have the
    same shape as the input patch. The consequence of this is that NaN values
    will appear at the start of the dimension. You can use
    [`patch.dropna`](`dascore.Patch.dropna`) to remove the NaN values.

    Second, the step parameter is equivalent applying to the output along the
    specified dimension. For example, if step=2 the output of the chosen
    dimension will be 1/2 of the input length.

    Here are a few examples to help illustrate how rolling works.

    Consider a patch with a simple 1D array in the dimension "time":
        [0, 1, 2, 3, 4, 5]
    If time = 2 * dt the output is
        [NaN, 0.5, 1.5, 2.5, 3.5, 4.5]
    If time = 3 * dt the output is
        [NaN, NaN, 1.0, 2.0, 3.0, 4.0]
    if time = 3 * dt and step = 2 * dt
        [NaN, 1.0, 3.0]
    if time = 3 * dt and step = 3 * dt
        [NaN, 2.0]
    """

    def _get_engine(step, engine, patch):
        """Get the engine."""
        engines = {"numpy": _NumpyPatchRoller, "pandas": _PandasPatchRoller}
        if cls := engines.get(engine):
            return cls
        if step < 10 and len(patch.dims) < 2:
            return _PandasPatchRoller
        return _NumpyPatchRoller

    # get window sizes in samples
    dim, axis, value = get_dim_value_from_kwargs(patch, kwargs)
    roll_hist = f"rolling({dim}={value}, step={step}, center={center}, engine={engine})"
    coord = patch.get_coord(dim)
    window = coord.get_sample_count(value, samples=samples)
    step = 1 if step is None else coord.get_sample_count(step, samples=samples)
    if window == 0 or step == 0:
        msg = "Window or step size can't be zero. Use any positive values."
        raise ParameterError(msg)
    cls = _get_engine(step, engine, patch)
    out = cls(
        patch=patch,
        window=window,
        step=step,
        dim=dim,
        axis=axis,
        center=center,
        roll_hist=roll_hist,
    )
    return out



"""Utilities for terra15."""
from __future__ import annotations

import numpy as np

import dascore as dc
from dascore.core import get_coord

# --- Getting format/version

DATA_ARRAY_NAMES = frozenset(["raw", "data"])
TIME_ARRAY_NAMES = frozenset(("timestamp", "time", "timestamps"))
OTHER_COORD_ARRAY_NAMES = frozenset(("channels", "distance"))

FILE_FORMAT_ATTR_NAMES = frozenset(("__format__", "file_format", "format"))
DEFAULT_ATTRS = frozenset(("CLASS", "PYTABLES_FORMAT_VERSION", "TITLE", "VERSION"))


def _maybe_trim_data(cm, data, kwargs):
    """Maybe use kwargs to trim data array."""
    new_cm, new_data = cm.select(array=data, **kwargs)
    return new_cm, new_data


def _get_attrs_coords_and_data(h5, snap, fiber_io):
    """Return attrs, coordinate manager, and data node."""
    attrs = h5.root._v_attrs
    attr_names = set(attrs._v_attrnames) - DEFAULT_ATTRS
    attr_dict = {x: getattr(attrs, x) for x in attr_names}
    attr_dict["file_version"] = fiber_io.version
    attr_dict["file_format"] = fiber_io.name
    cm, data = _get_cm_and_data(h5, snap, dims=attr_dict.get("dims"))
    attr_dict["dims"] = cm.dims
    return attr_dict, cm, data


def _get_coord(v, snap, name):
    """Get the coord values from a node."""
    if snap:
        start = v[0] if name != "time" else dc.to_datetime64(v[0])
        stop = v[-1] if name != "time" else dc.to_datetime64(v[-1])
        duration = stop - start
        step = duration / (len(v) - 1)
        coord = get_coord(min=start, max=stop + step, step=step)
        assert len(coord) == len(v)
    else:
        values = v[:] if name != "time" else dc.to_datetime64(v[:])
        coord = get_coord(values=values)
    return coord


def _fill_coords(coord_shape_dict, other_nodes, data_node):
    """
    Fill missing coordinate with "channel".

    This is needed because the foresee data on pubdas only specify time;
    we have to fill in channel number.
    """
    missing_shape = set(data_node.shape) - set(coord_shape_dict)
    assert len(missing_shape) == 1, "can only fill one missing coord."
    shape = next(iter(missing_shape))
    coord_shape_dict[shape] = "channel"
    other_nodes["channel"] = np.arange(shape)
    return other_nodes, coord_shape_dict


def _get_coords_and_dims(data_node, time_node, other_nodes, snap=True, dims=None):
    """Get dims tuple and coord dict."""
    if dims:
        dims = dims if not isinstance(dims, str) else dims.split(",")
    else:  # ascertain dims from shape
        can_guess_shape = len(data_node.shape) == len(set(data_node.shape))
        assert can_guess_shape, "Cant determine dims; shape values not unique!"
        assert len(time_node.shape) == 1, "time node has more than one dimension!"
        # get a dict of {coord_name: shape} for 1d coords.
        coord_shape_dict = {
            len(v): x for x, v in other_nodes.items() if len(v.shape) == 1
        }
        coord_shape_dict[len(time_node)] = "time"
        # need to fill some dims
        if len(coord_shape_dict) != len(data_node.shape):
            other_nodes, coord_shape_dict = _fill_coords(
                coord_shape_dict,
                other_nodes,
                data_node,
            )

        dims = tuple(coord_shape_dict[x] for x in data_node.shape)
    other_nodes["time"] = time_node
    coords = {i: _get_coord(v, snap=snap, name=i) for i, v in other_nodes.items()}
    return dims, coords


def _get_cm_and_data(h5, snap=False, dims=None):
    """Extract coordinate manager and data node."""
    array_names = {x.name for x in h5.list_nodes("/") if hasattr(x, "shape")}
    data_node_name = array_names & DATA_ARRAY_NAMES
    time_node_name = array_names & TIME_ARRAY_NAMES
    other_node_names = array_names - data_node_name - time_node_name

    assert len(data_node_name) == 1, f"{h5} doesn't have exactly one data node."
    assert len(time_node_name) == 1, f"{h5} doesn't have exactly one time node"

    data_node = getattr(h5.root, next(iter(data_node_name)))
    time_node = getattr(h5.root, next(iter(time_node_name)))
    other_nodes = {x: getattr(h5.root, x) for x in other_node_names}

    dims, coords = _get_coords_and_dims(data_node, time_node, other_nodes, snap, dims)
    return dc.core.get_coord_manager(coords, dims=dims), data_node


def _is_h5simple(h5):
    """Determine if open h5 file is simple H5."""
    has_arrays = _has_required_arrays(h5)
    version_ok = _no_format_or_simple_specified(h5)
    if has_arrays and version_ok:
        return True
    return False


def _has_required_arrays(h5):
    """Determine if h5 file has required arrays to be h5 simple."""
    array_names = {x.name for x in h5.list_nodes("/") if hasattr(x, "shape")}
    data_node = array_names & DATA_ARRAY_NAMES
    time_node = array_names & TIME_ARRAY_NAMES
    return bool(data_node) and bool(time_node)


def _no_format_or_simple_specified(h5):
    """Ensure no other format is specified, or that simpleH5 is."""
    attrs = h5.root._v_attrs
    attr_names = set(attrs._v_attrnames)
    file_format = attr_names & FILE_FORMAT_ATTR_NAMES
    format = getattr(attrs, next(iter(file_format))) if file_format else "h5simple"
    if format == "h5simple":
        return True
    return False



"""Module for reporting the version of dascore."""
from __future__ import annotations

from contextlib import suppress
from importlib.metadata import PackageNotFoundError, version

__version__ = "0.0.0"

# try to get version from installed metadata
with suppress(PackageNotFoundError):
    __version__ = version("dascore")

__last_version__ = ".".join(__version__.split(".")[:3])



"""Tests for FileSpool."""
from __future__ import annotations

from pathlib import Path

import numpy as np
import pandas as pd
import pytest

import dascore as dc
import dascore.examples
from dascore.clients.dirspool import DirectorySpool
from dascore.constants import ONE_SECOND
from dascore.exceptions import ParameterError
from dascore.io.core import PatchFileSummary
from dascore.utils.hdf5 import HDFPatchIndexManager
from dascore.utils.misc import register_func

DIRECTORY_SPOOLS = []


@pytest.fixture(scope="class")
@register_func(DIRECTORY_SPOOLS)
def dir_spool_index_out_of_order(random_spool, tmp_path_factory):
    """Create an index that isn't order chronologically."""
    path = tmp_path_factory.mktemp("out_of_order_index")
    spool = dc.spool(path)
    # sort patches by starttime
    patch_list = sorted(random_spool, key=lambda x: x.attrs.time_min)
    # write patches to disk out of order.
    patch_list[-1].io.write(path / "patch_3.h5", "dasdae")
    spool.update()
    patch_list[0].io.write(path / "patch_1.h5", "dasdae")
    spool.update()
    patch_list[1].io.write(path / "patch_2.h5", "dasdae")
    spool.update()
    return spool


@pytest.fixture(scope="class")
@register_func(DIRECTORY_SPOOLS)
def one_directory_spool(one_file_dir):
    """Create a directory with a single DAS file."""
    spool = DirectorySpool(one_file_dir)
    return spool.update()


@pytest.fixture(scope="class")
@register_func(DIRECTORY_SPOOLS)
def non_distance_dir_spool(tmp_path_factory):
    """Create a directory with a single DAS file."""
    # patch one simulates a patch that has time but no distance
    pa1 = dascore.examples.get_example_patch("random_das").rename_coords(
        distance="depth"
    )
    # patch2 has neither time nor distance but time in attrs.
    pa2 = (
        dascore.examples.get_example_patch("random_das")
        .update_attrs(time_min=pa1.attrs.time_max)
        .rename_coords(time="timey", distance="depth")
    )
    coord = pa2.get_coord("timey")
    pa2 = pa2.update_attrs(time_min=coord.min(), time_max=coord.max())
    spool = dc.spool([pa1, pa2])
    path = tmp_path_factory.mktemp("no_distance_spool")
    dascore.examples.spool_to_directory(spool, path)
    return dc.spool(path).update()


@pytest.fixture(scope="class", params=DIRECTORY_SPOOLS)
def directory_spool(request):
    """Meta fixture for getting all file spools."""
    return request.getfixturevalue(request.param)


class TestDirectorySpoolBasics:
    """Basic tests for the directory spool."""

    def test_isinstance(self, directory_spool):
        """Simply ensure expected type was returned."""
        assert isinstance(directory_spool, DirectorySpool)

    def test_selected_str(self, diverse_directory_spool):
        """Ensure select kwargs show up in str."""
        new = diverse_directory_spool.select(station="big_gaps")
        contents = new.get_contents()
        assert (contents["station"] == "big_gaps").all()


class TestDirectoryIndex:
    """Tests for returning summaries of all files in managed directory."""

    @pytest.fixture(scope="class")
    def basic_index_df(self, basic_file_spool):
        """Return the index file of the basic bank."""
        spool = basic_file_spool
        return spool.get_contents()

    def test_index_exists(self, basic_file_spool):
        """An index should be returned."""
        assert basic_file_spool.indexer.index_path.exists()

    def test_index_len(self, basic_index_df, two_patch_directory):
        """An index should be returned."""
        spool = dc.spool(two_patch_directory)
        spool.indexer.index_path.unlink()
        df = spool.update().get_contents()
        bank_paths = list(Path(two_patch_directory).rglob("*hdf5"))
        assert isinstance(df, pd.DataFrame)
        assert len(bank_paths) == len(df)

    def test_index_columns(self, basic_index_df):
        """Ensure expected columns show up in the index."""
        schema_fields = list(PatchFileSummary.model_fields)
        assert set(basic_index_df).issuperset(schema_fields)

    def test_patches_extracted(self, basic_file_spool):
        """Ensure the patches can be extracted."""
        index = basic_file_spool.get_contents()
        patches = [x for x in basic_file_spool]
        assert len(index) == len(patches)
        for patch in patches:
            assert isinstance(patch, dc.Patch)

    def test_str_(self, basic_file_spool):
        """Ensure the filespool has a useful str/repr."""
        out = str(basic_file_spool)
        # ensure the default str is not used.
        assert "object at" not in out

    def test_long_tags(self, random_patch, tmp_path):
        """Ensure a long tag still works."""
        new = random_patch.update_attrs(tag="hey" * 15)
        path = tmp_path / "test.h5"
        new.io.write(path, "dasdae")
        spool = dc.spool(path).update()
        isinstance(spool, dc.BaseSpool)

    def test_specify_index_path(self, random_patch, tmp_path_factory):
        """Ensure an external path can be specified for the index. See #129."""
        bank_path = tmp_path_factory.mktemp("bank")
        index_path = tmp_path_factory.mktemp("index") / "index.h5"
        random_patch.io.write(bank_path / "contents.h5", "dasdae")
        spool1 = dc.spool(bank_path, index_path=index_path)
        spool1.update()
        # ensure the index was created in the expected place
        assert spool1.indexer.index_path == index_path
        # ensure the default index file was not written
        default_index_path = bank_path / spool1.indexer._index_name
        assert not default_index_path.exists()
        # future banks should remember this path.
        spool2 = dc.spool(bank_path)
        assert spool2.indexer.index_path == spool1.indexer.index_path
        # next ensure the index path is used
        spool3 = dc.spool(bank_path, index_path=index_path)
        df = spool3.get_contents()
        assert len(df) == 1
        patch = spool3[0]
        assert isinstance(patch, dc.Patch)
        assert not default_index_path.exists()

    def test_nested_directories(self, diverse_spool, tmp_path_factory):
        """Ensure files in nested directories work up to 3 levels."""
        # split the spool into 3
        sp_len = len(diverse_spool)
        num = 3
        spools = [
            diverse_spool[int((x / num) * sp_len) : int(((x + 1) / num) * sp_len)]
            for x in range(num)
        ]
        # write each group to a different sub path
        base_path = tmp_path_factory.mktemp("nested_dir")
        path = base_path
        for num, spool in enumerate(spools):
            path = path / f"sub_{num}"
            path.mkdir(exist_ok=True, parents=True)
            dascore.examples.spool_to_directory(spool, path)
        df = dc.spool(base_path).update().get_contents()
        # ensure each sub-directory is represented
        paths = df["path"]
        assert any(paths.str.startswith("sub_0"))
        assert any(paths.str.startswith("sub_0/sub_1"))
        assert any(paths.str.startswith("sub_0/sub_1/sub_2"))


class TestSelect:
    """tests for subselecting data."""

    @pytest.fixture(scope="class")
    def spool_tag(self, basic_file_spool):
        """Return a string of a tag in the basic_file_spool."""
        contents = basic_file_spool.get_contents()
        tag = contents.loc[contents["tag"].astype(bool), "tag"].iloc[0]
        return tag

    def test_subselect_trims_patches(self, basic_file_spool):
        """Ensure sub-selecting trims start/end times on df and output patches."""
        current = basic_file_spool.get_contents()
        new_min = current["time_min"].min() + np.timedelta64(2, "s")
        new_max = current["time_max"].max() - np.timedelta64(2, "s")
        spool = basic_file_spool.select(time=(new_min, new_max))
        # the limits of rows which were intersected should have been trimmed.
        df = spool.get_contents()
        assert (df["time_min"] >= new_min).all()
        assert (df["time_max"] <= new_max).all()
        # as well as the patches produced
        for patch in spool:
            assert patch.attrs["time_min"] >= new_min
            assert patch.attrs["time_max"] <= new_max

    def test_sub_select_tag_equals(self, basic_file_spool, spool_tag):
        """Ensure selecting stations works."""
        new = basic_file_spool.select(tag=spool_tag)
        new_contents = new.get_contents()
        assert (new_contents["tag"] == spool_tag).all()

    def test_is_in_tag(self, basic_file_spool, spool_tag):
        """Ensure tags can also be selected from a collection."""
        tag_collection = {spool_tag, "bob", "bill"}
        out = basic_file_spool.select(tag=tag_collection).get_contents()
        assert out["tag"].isin(tag_collection).all()

    def test_multiple_selects(self, diverse_directory_spool):
        """Ensure selects can be stacked."""
        spool = diverse_directory_spool
        contents = spool.get_contents()
        duration = contents["time_max"] - contents["time_min"]
        new_max = (contents["time_min"] + duration.mean() / 2).median()
        out = (
            spool.select(network="das2").select(tag="ran*").select(time=(None, new_max))
        )
        assert len(out) > 0
        # first check content dataframe
        new_content = out.get_contents()
        assert len(new_content) == len(out)
        assert (new_content["network"] == "das2").all()
        assert (new_content["tag"].str.startswith("ran")).all()
        assert (new_content["time_max"] <= new_max).all()
        # then check patches
        for patch in out:
            attrs = patch.attrs
            assert attrs["network"] == "das2"
            assert attrs["tag"].startswith("ran")
            assert attrs["time_max"] <= new_max
        # ensure raises when selecting off the end of the spool
        with pytest.raises(IndexError):
            out[len(new_content)]

    def test_select_time_tuple_with_string(self, basic_file_spool):
        """Ensure time tuples with strings still work."""
        time_str = "2017-09-18T00:00:04"
        dt = np.datetime64(time_str)
        spool1 = basic_file_spool.select(time=(None, dt))
        spool2 = basic_file_spool.select(time=(None, time_str))
        for pa1, pa2 in zip(spool1, spool2):
            assert pa1.attrs["time_max"] == pa2.attrs["time_max"]

    def test_select_non_zero_index(self, diverse_directory_spool):
        """
        A Bug caused the contents of the source dataframe to have
        non-zero based indices, thus spools didn't work.
        """
        contents = diverse_directory_spool.get_contents()
        end_time = contents["time_max"].min()
        sub = diverse_directory_spool.select(
            time=(None, end_time),
            distance=(100, 200),
        )
        assert len(sub) == 1
        patch = sub[0]
        assert isinstance(patch, dc.Patch)

    def test_nice_error_message_bad_select(self, diverse_directory_spool):
        """Ensure a nice error message is raised for bad filter param."""
        with pytest.raises(ParameterError, match="must be a length 2 sequence"):
            _ = diverse_directory_spool.select(time=(None, None, None))[0]

    def test_select_correct_history_str(self, diverse_directory_spool):
        """Ensure no history string is added for selecting. See #142/#147."""
        spool = diverse_directory_spool
        t1 = spool[0].attrs.time_min
        dt = spool[0].attrs.time_step
        selected_spool = spool.select(time=(t1, t1 + 30 * dt))
        patch = selected_spool[0]
        history = patch.attrs.history
        assert len(history) <= 1


class TestBasicChunk:
    """Tests for chunking filespool."""

    @pytest.fixture(scope="class")
    def dir_spool_1_dim_patches(self, memory_spool_dim_1_patches, tmp_path_factory):
        """Create a directory with patches that have 1 dim in time."""
        path = tmp_path_factory.mktemp("dir_spool_1_dim_patches")
        out = dc.examples.spool_to_directory(memory_spool_dim_1_patches, path)
        return dc.spool(out).update()

    def test_directory_path_doesnt_change(self, one_file_directory_spool):
        """Chunking shouldn't change the path to the managed directory."""
        out = one_file_directory_spool.chunk(time=1)
        assert out.spool_path == one_file_directory_spool.spool_path

    def test_chunk_doesnt_modify_original(self, one_file_directory_spool):
        """Chunking shouldn't modify original spool or its dfs."""
        spool = one_file_directory_spool
        contents_before_chunk = spool.get_contents()
        _ = spool.chunk(time=2)
        contents_after_chunk = spool.get_contents()
        assert contents_before_chunk.equals(contents_after_chunk)

    def test_sub_chunk(self, one_file_directory_spool):
        """Ensure the patches can be subdivided."""
        spool = one_file_directory_spool
        contents = spool.get_contents()
        durations = contents["time_max"] - contents["time_min"]
        new_t_delta = (durations / 4).max()
        new_spool = spool.chunk(time=new_t_delta, keep_partial=True)
        # Ensure there are exactly 4x as many patches in spool after chunk
        new_contents = new_spool.get_contents()
        assert len(new_contents) == 4 * len(spool)
        # Ensure each spool can be iterated
        patch_list = list(new_spool)
        for patch in patch_list:
            assert isinstance(patch, dc.Patch)

    def test_merge_1_dim_patches(self, dir_spool_1_dim_patches):
        """Ensure patches with one sample in time can be merged."""
        spool = dir_spool_1_dim_patches
        new = spool.chunk(time=None)
        assert len(new) == 1
        patch = new[0]
        content = spool.get_contents()
        assert patch.attrs.time_min == content["time_min"].min()
        assert patch.attrs.time_max == content["time_max"].max()
        assert patch.attrs.time_step == spool[0].attrs.time_step

    def test_chunk_out_of_order_index(self, dir_spool_index_out_of_order):
        """Ensure when the index isn't ordered chunk can still work."""
        spool = dir_spool_index_out_of_order
        time = 4.25
        chunk = spool.chunk(time=time)
        for patch in chunk:
            assert isinstance(patch, dc.Patch)
            dur = (patch.attrs.time_max - patch.attrs.time_min) / ONE_SECOND
            diff = np.abs(dur - time)
            # because we try to avoid overlaps, the segments can be up to 2
            # samples shorter than what was asked for. Maybe revisit this?
            assert diff <= 2 * (patch.attrs.time_step / ONE_SECOND)


class TestGetContents:
    """Tests for getting the contents of the spool."""

    def test_str_columns_in_dataframe(self, diverse_directory_spool):
        """Ensure all the string columns are in index."""
        df = diverse_directory_spool.get_contents()
        expected = HDFPatchIndexManager._min_itemsize
        assert set(df.columns).issuperset(set(expected))


class TestIndexing:
    """Tests for indexing directory spool."""

    def test_slice_to_start(self, diverse_directory_spool):
        """Ensure a slice returns a subspool (shouldn't load data)."""
        out = diverse_directory_spool[0:2]
        assert isinstance(out, out.__class__)

    def test_slice_to_end(self, diverse_directory_spool):
        """Ensure a slice from the end returns a subspool."""
        out = diverse_directory_spool[-2:]
        assert isinstance(out, out.__class__)

    def test_sliced_spool_has_indexer(self, diverse_directory_spool):
        """Ensure the sliced spool still has its indexer."""
        out = diverse_directory_spool[1:3]
        assert hasattr(out, "indexer")
        assert out.indexer.path == diverse_directory_spool.indexer.path
        # ensure we can still load patches from sliced dirspool
        assert isinstance(out[0], dc.Patch)

    def test_chunked_sliced_spool_index(self, diverse_directory_spool):
        """Ensure chunked sliced spool can still be indexed and patches loaded."""
        out = diverse_directory_spool.chunk(time=4)
        middle_index = len(out) // 2
        sub = out[middle_index : middle_index + 3]
        for ind in range(len(sub)):
            patch = sub[ind]
            assert isinstance(patch, dc.Patch)


class TestFileSpoolIntegrations:
    """Small integration tests for the file spool."""

    def test_one(self, diverse_spool_directory):
        """Small integration test with diverse spool."""
        network = "das2"
        endtime = np.datetime64("2022-01-01")
        duration = 3
        spool = (
            dc.spool(diverse_spool_directory)
            .select(network=network)  # sub-select das2 network
            .select(time=(None, endtime))  # unselect anything after 2022
            .chunk(time=duration, overlap=0.5)  # change the chunking of the patches
        )
        for patch in spool:
            assert isinstance(patch, dc.Patch)
            attrs = patch.attrs
            assert attrs["network"] == network
            assert attrs["time_max"] <= endtime
            patch_duration = (attrs["time_max"] - attrs["time_min"]) / ONE_SECOND
            diff = patch_duration - duration
            assert abs(diff) <= 1.5 * attrs["time_step"] / ONE_SECOND

    def test_chunk_select(self, dir_spool_index_out_of_order):
        """Ensure chunking can be performed first, then selecting."""
        # get start/endtimes to encompass the last half of the first patch.
        # and the first half of the second patch.
        df = dir_spool_index_out_of_order.get_contents().sort_values("time_min")
        time = (df["time_max"] - df["time_min"]) / 2 + df["time_min"]
        time_tup = (time.iloc[0], time.iloc[1])
        # merge, then select, should still work.
        merged = dir_spool_index_out_of_order.chunk(time=...)
        assert len(merged) == 1
        select = merged.select(time=time_tup)
        assert len(select) == 1

    def test_doc_example(self, all_examples_spool):
        """Tests for quickstart."""
        spool = all_examples_spool.update()
        assert isinstance(spool, dc.BaseSpool)

    def test_patch_no_distance_coord(self, non_distance_dir_spool):
        """Ensure patches without distance coords still work."""
        # str should work
        assert str(non_distance_dir_spool)
        contents = non_distance_dir_spool.get_contents()
        assert len(contents) == 2
        for patch in non_distance_dir_spool:
            assert isinstance(patch, dc.Patch)

    def test_select_non_distance(self, non_distance_dir_spool):
        """We should be able to select on non-time/distance coords."""
        spool = non_distance_dir_spool
        depth_tup = (150, 250)
        selected_spool = spool.select(depth=depth_tup)
        for patch in selected_spool:
            # ensure depth has been trimmed.
            coord = patch.get_coord("depth")
            assert coord.min() >= depth_tup[0]
            assert coord.max() <= depth_tup[1]



"""Utilities for working with the Patch class."""
from __future__ import annotations

import functools
import inspect
import sys
import warnings
from collections.abc import Callable, Mapping, Sequence
from typing import Any, Literal

import numpy as np
import pandas as pd

import dascore as dc
from dascore.constants import FLOAT_PRECISION, PatchType, SpoolType, dascore_styles
from dascore.core.attrs import combine_patch_attrs
from dascore.core.coordmanager import merge_coord_managers
from dascore.exceptions import (
    CoordDataError,
    ParameterError,
    PatchAttributeError,
    PatchDimError,
)
from dascore.units import get_quantity
from dascore.utils.misc import all_diffs_close_enough, get_middle_value, iterate
from dascore.utils.time import to_float

attr_type = dict[str, Any] | str | Sequence[str] | None


def _format_values(val):
    """String formatting for values for history string."""
    if isinstance(val, list | tuple):
        out = ",".join(_format_values(x) for x in val)
        out = f"({out})" if isinstance(val, tuple) else f"[{out}]"
    elif isinstance(val, np.ndarray):
        # make sure numpy strings arent too long!
        out = np.array2string(
            val,
            precision=FLOAT_PRECISION,
            threshold=dascore_styles["patch_history_array_threshold"],
        )
    else:
        out = str(val)
    return out


def _func_and_kwargs_str(func: Callable, patch, *args, **kwargs) -> str:
    """Get a str rep of the function and input args."""
    callargs = inspect.getcallargs(func, patch, *args, **kwargs)
    callargs.pop("patch", None)
    callargs.pop("self", None)
    kwargs_ = callargs.pop("kwargs", {})
    arguments = []
    arguments += [
        f"{k}={_format_values(v)!r}" for k, v in callargs.items() if v is not None
    ]
    arguments += [
        f"{k}={_format_values(v)!r}" for k, v in kwargs_.items() if v is not None
    ]
    arguments.sort()
    out = f"{func.__name__}("
    if arguments:
        out += f"{','.join(arguments)}"
    return out + ")"


def _get_history_str(patch: PatchType, func, *args, _history="full", **kwargs) -> str:
    """
    Log history of a function being called on a Patch object.

    Parameters
    ----------
    patch
        The patch which will track history.
    func
        A callable which takes the patch as the first argument.
    *args
        The arguments passed to the function.
    _history
        String specifying how the history is to be recorded.
    **kwargs
        kwargs for func.
    """
    if _history is None:
        return None
    if _history == "full":
        history_str = _func_and_kwargs_str(func, patch, *args, **kwargs)
    else:
        history_str = str(func.__name__)
    return history_str


def check_patch_dims(
    patch: PatchType, required_dims: tuple[str, ...] | None
) -> PatchType:
    """
    Check if a patch object has required dimensions, else raise PatchDimError.

    Parameters
    ----------
    patch
        The input patch
    required_dims
        A tuple of required dimensions.
    """
    if required_dims is None:
        return patch
    dim_set = set(patch.dims)
    missing_dims = set(required_dims) - set(patch.dims)
    if not set(required_dims).issubset(dim_set):
        msg = f"patch is missing required dims: {tuple(missing_dims)}"
        raise PatchDimError(msg)
    return patch


def check_patch_attrs(patch: PatchType, required_attrs: attr_type) -> PatchType:
    """
    Check for expected attributes.

    Parameters
    ----------
    patch
        The patch to validate
    required_attrs
        The expected attrs. Can be a sequence or mapping. If sequence, only
        check that attrs exist. If mapping also check that values are equal.
    """
    if required_attrs is None:
        return patch
    # test that patch attr mapping is equal
    patch_attrs = patch.attrs
    if isinstance(required_attrs, Mapping):
        sub = {i: patch_attrs[i] for i in required_attrs}
        if sub != dict(required_attrs):
            msg = f"Patch's attrs {sub} are not required attrs: {required_attrs}"
            raise PatchAttributeError(msg)
    else:
        missing = set(required_attrs) - set(dict(patch.attrs))
        if missing:
            msg = f"Patch is missing the following attributes: {missing}"
            raise PatchAttributeError(msg)
    return patch


def patch_function(
    required_dims: tuple[str, ...] | Callable | None = None,
    required_attrs: attr_type = None,
    history: Literal["full", "method_name", None] = "full",
):
    """
    Decorator to mark a function as a patch method.

    Parameters
    ----------
    required_dims
        A tuple of dimensions which must be found in the Patch.
    required_attrs
        A dict of attributes which must be found in the Patch and whose
        values must be equal to those provided.
    history
        Specifies how to track history on Patch.
            Full - Records function name and str version of input arguments.
            method_name - Only records method name. Useful if args are long.
            None - Function call is not recorded in history attribute.

    Examples
    --------
    1. A patch method which requires dimensions (time, distance)
    >>> import dascore as dc
    >>> @dc.patch_function(required_dims=('time', 'distance'))
    ... def do_something(patch):
    ...     ...   # raises a PatchCoordsError if patch doesn't have time,
    ...     #  distance

    2. A patch method which requires an attribute 'data_type' == 'DAS'
    >>> import dascore as dc
    >>> @dc.patch_function(required_attrs={'data_type': 'DAS'})
    ... def do_another_thing(patch):
    ...     ...  # raise PatchAttributeError if patch doesn't have attribute
    ...     # called "data_type" or its values is not equal to "DAS".

    Notes
    -----
    The original function can still be accessed with the .func attribute.
    This may be useful for avoiding calling the patch_func machinery
    multiple times from within another patch function.
    """

    def _wrapper(func):
        @functools.wraps(func)
        def _func(patch, *args, **kwargs):
            check_patch_dims(patch, required_dims)
            check_patch_attrs(patch, required_attrs)
            hist_str = _get_history_str(patch, func, *args, _history=history, **kwargs)
            out: PatchType = func(patch, *args, **kwargs)
            # attach history string. Consider something a bit less hacky.
            if hist_str and hasattr(out, "attrs"):
                hist = list(out.attrs.history)
                hist.append(hist_str)
                out = out.update_attrs(history=hist)
            return out

        _func.func = func  # attach original function
        _func.__wrapped__ = func

        return _func

    if callable(required_dims):  # the decorator is used without parens
        return patch_function()(required_dims)

    return _wrapper


def patches_to_df(
    patches: Sequence[PatchType] | SpoolType | pd.DataFrame,
) -> pd.DataFrame:
    """
    Return a dataframe.

    Parameters
    ----------
    patches
        A sequence of :class:`dascore.Patch`

    Returns
    -------
    A dataframe with the attrs of each patch converted to a columns
    plus a field called 'patch' which contains a reference to the patches.
    """
    if hasattr(patches, "_df"):
        df = patches._df
    # Handle spool case
    elif hasattr(patches, "get_contents"):
        df = patches.get_contents()
    elif isinstance(patches, pd.DataFrame):
        df = patches
    else:
        df = pd.DataFrame([x.flat_dump() for x in scan_patches(patches)])
        if df.empty:  # create empty df with appropriate columns
            cols = list(dc.PatchAttrs().model_dump())
            df = pd.DataFrame(columns=cols).assign(patch=None, history=None)
        else:  # else populate with patches and concat history
            history = df["history"].apply(lambda x: ",".join(x))
            df = df.assign(patch=patches, history=history)
    # Ensure history is in df
    if "history" not in df.columns:
        df = df.assign(history="")
    if "patch" not in df.columns:
        df["patch"] = [x for x in patches]
    return df


def merge_patches(
    patches: Sequence[PatchType] | pd.DataFrame | SpoolType,
    dim: str = "time",
    check_history: bool = True,
    tolerance: float = 1.5,
) -> Sequence[PatchType]:
    """
    Merge all compatible patches in spool or patch list together.

    Parameters
    ----------
    patches
        A sequence of patches to merge (if compatible)
    dim
        The dimension along which to merge
    check_history
        If True, only merge patches with common history. This will, for
        example, prevent merging filtered and unfiltered data together.
    tolerance
        The upper limit of a gap to tolerate in terms of the sampling
        along the desired dimension. e.g., the default value means any patches
        with gaps <= 1.5 * dt will be merged.
    """
    msg = (
        "merge_patches is deprecated. Use spool.chunk instead. "
        "For example, to merge a list of patches you can use: "
        "dascore.spool(patch_list).chunk(time=None) to merge on the time "
        "dimension."
    )
    warnings.warn(msg, DeprecationWarning, stacklevel=2)
    return dc.spool(patches).chunk(**{dim: None}, tolerance=tolerance)


def _force_patch_merge(patch_dict_list, merge_kwargs, **kwargs):
    """
    Force a merge of the patches along a dimension.

    This function is used in conjunction with `spool.chunk`, which
    does all the compatibility checks beforehand.
    """

    def _get_merge_col(df):
        dims = df["dims"].unique()
        assert len(dims) == 1
        dims = dims[0].split(",")
        dims_vary = pd.Series({x: False for x in dims})
        for dim in dims:
            cols = [f"{dim}_min", f"{dim}_max", f"{dim}_step"]
            vals = df[cols].values
            vals_eq = vals == vals[[0], :]
            vals_null = pd.isnull(vals)
            columns_equal = (vals_eq | vals_null).all(axis=1)
            dims_vary[dim] = not np.all(columns_equal)
        assert dims_vary.sum() <= 1, "Only one dimension can vary for forced merge"
        if not dims_vary.any():  # the case of complete overlap.
            return None
        return dims_vary[dims_vary].index[0]

    def _maybe_step(df, dim):
        """Get the expected step if all steps are close, else None."""
        col = df[f"{dim}_step"].values
        if all_diffs_close_enough(col):
            return get_middle_value(col)
        return None

    def _get_new_coord(df, merge_dim, coords):
        """Get new coordinates, also validate anticipated sampling."""
        new_coord = merge_coord_managers(coords, dim=merge_dim)
        expected_step = _maybe_step(df, merge_dim)
        if not pd.isnull(expected_step):
            new_coord = new_coord.snap(merge_dim)[0]
            # TODO slightly different dt can be produced, let pass for now
            # need to think more about how the merging should work.
        return new_coord

    df = pd.DataFrame(patch_dict_list)
    merge_dim = _get_merge_col(df)
    merge_kwargs = merge_kwargs if merge_kwargs is not None else {}
    if merge_dim is None:  # nothing to merge, complete overlap
        return [patch_dict_list[0]]
    dims = df["dims"].iloc[0].split(",")
    # get patches, ensure they are oriented the same.
    patches = [x.transpose(*dims) for x in df["patch"]]
    axis = patches[0].dims.index(merge_dim)
    # get data, coords, attrs for merging patch together.
    datas = [x.data for x in patches]
    coords = [x.coords for x in patches]
    attrs = [x.attrs for x in patches]
    new_data = np.concatenate(datas, axis=axis)
    new_coord = _get_new_coord(df, merge_dim, coords)
    coord = new_coord.coord_map[merge_dim] if merge_dim in dims else None
    new_attrs = combine_patch_attrs(attrs, merge_dim, coord=coord, **merge_kwargs)
    patch = dc.Patch(data=new_data, coords=new_coord, attrs=new_attrs, dims=dims)
    new_dict = {"patch": patch}
    return [new_dict]


def scan_patches(patches: PatchType | Sequence[PatchType]) -> list[dc.PatchAttrs]:
    """
    Scan a sequence of patches and return a list of summaries.

    The summary dicts have the following fields:
        {fields}

    Parameters
    ----------
    patches
        A single patch or a sequence of patches.
    """
    if isinstance(patches, dc.Patch):
        patches = [patches]  # make sure we have an iterable
    out = [pa.attrs for pa in patches]
    return out


def get_start_stop_step(patch: PatchType, dim):
    """Convenience method for getting start, stop, step for a given coord."""
    assert dim in patch.dims, f"{dim} is not in Patch dimensions of {patch.dims}"
    coord = patch.get_coord(dim)
    start = coord.min()
    stop = coord.max()
    step = coord.step
    return start, stop, step


def get_default_patch_name(patch):
    """Generates the name of the node."""

    def _format_datetime64(dt):
        """Format the datetime string in a sensible way."""
        out = str(np.datetime64(dt).astype("datetime64[ns]"))
        return out.replace(":", "_").replace("-", "_").replace(".", "_")

    attrs = patch.attrs
    start = _format_datetime64(attrs.get("time_min", ""))
    end = _format_datetime64(attrs.get("time_max", ""))
    net = attrs.get("network", "")
    sta = attrs.get("station", "")
    tag = attrs.get("tag", "")
    return f"DAS__{net}__{sta}__{tag}__{start}__{end}"


def get_dim_value_from_kwargs(patch, kwargs):
    """
    Assert that kwargs contain one value and it is a dimension of patch.

    Several patch functions allow passing values via kwargs which are dimension
    specific. This function allows for some sane validation of such functions.

    Return the name of the dimension, its axis position, and its value.
    """
    dims = patch.dims
    overlap = set(dims) & set(kwargs)
    if len(kwargs) != 1 or not overlap:
        msg = (
            "You must use exactly one dimension name in kwargs. "
            f"You passed the following kwargs: {kwargs} to a patch with "
            f"dimensions {patch.dims}"
        )
        raise PatchDimError(msg)
    dim = next(iter(overlap))
    axis = dims.index(dim)
    return dim, axis, kwargs[dim]


def get_multiple_dim_value_from_kwargs(patch, kwargs):
    """Get multiple dim, axis and values from kwargs."""
    dims = patch.dims
    overlap = set(dims) & set(kwargs)
    if not overlap:
        msg = "You must specify one or more dimension in keyword args."
        raise ParameterError(msg)
    out = {}
    for dim in overlap:
        axis = patch.dims.index(dim)
        out[dim] = {"dim": dim, "axis": axis, "value": kwargs[dim]}
    return out


def get_dim_sampling_rate(patch: PatchType, dim: str) -> float:
    """
    Get sampling rate, as a float from sampling period along a dimension.

    Parameters
    ----------
    patch
        The imput patch.
    dim
        Dimension to extract.

    Raises
    ------
    [CoordDataError](`dascore.exceptions.CoordDataError`) if patch is not
    evenly sampled along desired dimension.
    """
    d_dim = patch.coords.coord_map[dim].step
    if isinstance(d_dim, np.timedelta64):
        d_dim = d_dim / np.timedelta64(1, "s")
    if pd.isnull(d_dim):
        # get the name of the calling function
        calling_function = inspect.getframeinfo(sys._getframe(1))[2]
        msg = (
            f"Patch coordinate {dim} is not evenly sampled as required by "
            f"{calling_function}. This can be fixed with Patch.snap or "
            f"Patch.extrapolate. "
        )
        raise CoordDataError(msg)
    return 1.0 / d_dim


def _get_data_units_from_dims(patch, dims, operator):
    """Get new data units from some operation on dimensions."""
    if (data_units := get_quantity(patch.attrs.data_units)) is None:
        return
    dim_units = None
    for dim in iterate(dims):
        dim_unit = get_quantity(patch.get_coord(dim).units)
        if dim_unit is None:
            continue
        dim_units = dim_unit if dim_units is None else dim_unit * dim_units
    if dim_units is not None:
        data_units = operator(data_units, dim_units)
    return data_units


def _get_dx_or_spacing_and_axes(
    patch,
    dim,
    require_sorted=True,
    require_evenly_spaced=False,
) -> tuple[tuple[float | np.ndarray, ...], tuple[int, ...]]:
    """
    Return dx (spacing) or values for a list of dims and corresponding axes.

    Parameters
    ----------
    patch
        The input patch
    dim
        The dimension name or sequence of such
    require_sorted
        If True, raise an error if all requested dimensions are not sorted.
    require_evenly_spaced
        If True, raise an error if all requested dimensions are not evenly sampled.
    """
    dims = iterate(dim if dim is not None else patch.dims)
    out = []
    axes = []
    for dim_ in dims:
        coord = patch.get_coord(
            dim_,
            require_sorted=require_sorted,
            require_evenly_sampled=require_evenly_spaced,
        )
        if coord.evenly_sampled:
            val = coord.step
        else:
            val = coord.data
        # need to convert val to float so datetimes work
        out.append(to_float(val))
        axes.append(patch.dims.index(dim_))

    return tuple(out), tuple(axes)



"""Utilities for TDMS format."""
from __future__ import annotations

import datetime
import mmap
import os
import struct

import numpy as np

from dascore.core.attrs import PatchAttrs
from dascore.core.coords import get_coord
from dascore.utils.time import to_datetime64, to_timedelta64

DEFAULT_ATTRS = set(PatchAttrs.model_fields)


def type_not_supported(vargin):
    """Function raises a NotImplementedException."""
    raise NotImplementedError("Reading of this tdsDataType is not implemented")


# Enum mapping TDM data types to description string, numpy type where exists
# See Ref[2] for enum values
TDS_DATA_TYPE = dict(
    {
        0x00: "void",  # tdsTypeVoid
        0x01: "int8",  # tdsTypeI8
        0x02: "int16",  # tdsTypeI16
        0x03: "int32",  # tdsTypeI32
        0x04: "int64",  # tdsTypeI64
        0x05: "uint8",  # tdsTypeU8
        0x06: "uint16",  # tdsTypeU16
        0x07: "uint32",  # tdsTypeU32
        0x08: "uint64",  # tdsTypeU64
        0x09: "float32",  # tdsTypeSingleFloat
        0x0A: "float64",  # tdsTypeDoubleFloat
        0x0B: "float128",  # tdsTypeExtendedFloat
        0x19: "singleFloatWithUnit",  # tdsTypeSingleFloatWithUnit
        0x1A: "doubleFloatWithUnit",  # tdsTypeDoubleFloatWithUnit
        0x1B: "extendedFloatWithUnit",  # tdsTypeExtendedFloatWithUnit
        0x20: "str",  # tdsTypeString
        0x21: "bool",  # tdsTypeBoolean
        0x44: "datetime",  # tdsTypeTimeStamp
        0xFFFFFFFF: "raw",  # tdsTypeDAQmxRawData
    }
)

# Function mapping for reading TDMS data types
TDS_READ_VAL = dict(
    {
        "void": lambda f: None,  # tdsTypeVoid
        "int8": lambda f: struct.unpack("<b", f.read(1))[0],
        "int16": lambda f: struct.unpack("<h", f.read(2))[0],
        "int32": lambda f: struct.unpack("<i", f.read(4))[0],
        "int64": lambda f: struct.unpack("<q", f.read(8))[0],
        "uint8": lambda f: struct.unpack("<B", f.read(1))[0],
        "uint16": lambda f: struct.unpack("<H", f.read(2))[0],
        "uint32": lambda f: struct.unpack("<I", f.read(4))[0],
        "uint64": lambda f: struct.unpack("<Q", f.read(8))[0],
        "float32": lambda f: struct.unpack("<f", f.read(4))[0],
        "float64": lambda f: struct.unpack("<d", f.read(8))[0],
        "float128": type_not_supported,
        "singleFloatWithUnit": type_not_supported,
        "doubleFloatWithUnit": type_not_supported,
        "extendedFloatWithUnit": type_not_supported,
        "str": lambda f: f.read(struct.unpack("<i", f.read(4))[0]),
        "bool": lambda f: struct.unpack("<?", f.read(1))[0],
        "datetime": lambda f: parse_time_stamp(
            struct.unpack("<Q", f.read(8))[0], struct.unpack("<q", f.read(8))[0]
        ),
        "raw": type_not_supported,
    }
)

DECIMATE_MASK = 0b00100000
FILEINFO_NAMES = (
    "file_tag",
    "toc",
    "version",
    "next_segment_offset",
    "raw_data_offset",
)


def parse_time_stamp(fractions, seconds):
    """
    Convert time TDMS time representation to datetime
    fractions   -- fractional seconds (2^-64)
    seconds     -- The number of seconds since 1/1/1904
    @rtype : datetime.datetime.
    """
    if fractions is not None and seconds is not None and fractions + seconds > 0:
        return datetime.timedelta(0, fractions * 2**-64 + seconds) + datetime.datetime(
            1904, 1, 1
        )
    else:
        return None


def _get_version_str(tdms_file, LEAD_IN_LENGTH=28) -> str:
    """Return True if we have the right file type."""
    lead_in = tdms_file.read(LEAD_IN_LENGTH)
    # lead_in is 28 bytes:
    # [string of length 4][int32][int32][int64][int64]
    fields = struct.unpack("<4siiQQ", lead_in)
    # TODO: validate file
    if fields[0].decode() in "TDSm":
        version_str = str(fields[2])
        return version_str
    else:
        return False


def _get_time_coord(attrs, num_samps):
    """Get the time array for the file."""
    dt = to_timedelta64(1 / attrs["SamplingFrequency[Hz]"])
    t_min = to_datetime64(str(attrs["GPSTimeStamp"]))
    # Note: Previously this was:
    # out["time_min"] + np.timedelta64(
    t_max = t_min + dt * (num_samps - 1)
    coord = get_coord(start=t_min, stop=t_max + dt, step=dt, units="s")
    return coord


def _get_default_attrs(tdms_file, attrs=None):
    """Return the required/default attributes which can be fetched from attributes."""
    all_attrs = attrs if attrs is not None else _get_all_attrs(tdms_file)[0]
    # cull attributes to only include defaults (TODO: think about why?)
    out = {
        default_attr: all_attrs[default_attr]
        for default_attr in DEFAULT_ATTRS
        if default_attr in all_attrs
    }
    return out


def _read_attr(tdms_file):
    """
    Read a single property from the TDMS file.
    Return the name, type and value of the property as a list.
    """
    # Read length of object path:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    # Read property name and type:
    name, data_type = struct.unpack(f"<{var}si", tdms_file.read(var + 4))
    # Lookup function to read and parse property value based on type:
    value = TDS_READ_VAL[TDS_DATA_TYPE[data_type]](tdms_file)
    name = name.decode()
    if data_type == 32:
        value = value.decode()

    return name, value  # data_type, value


def _get_distance_coord(attr):
    """Get distance coordinate from attribute."""
    # Note: some TDMS files actually have "Start Distance (m)" and
    # "Stop Distance (m)" fields, but not all. These also don't really
    # match the distance calculated below. We need to figure out why and what
    # is the correct way to do this, but this seems safe for now.
    multiplier = attr["Fibre Length Multiplier"]
    total_length = attr["MeasureLength[m]"] * multiplier
    start = attr["StartPosition[m]"] + total_length
    step = attr["SpatialResolution[m]"] * multiplier
    stop = start + (attr["n_channels"]) * step
    d_coord = get_coord(start=start, stop=stop, step=step, units="m")
    return d_coord


def _get_all_attrs(tdms_file, LEAD_IN_LENGTH=28):
    """Return all the attributes which can be fetched from attributes."""
    # read leadin infomation into fileinfo
    lead_in = tdms_file.read(LEAD_IN_LENGTH)
    # lead_in is 28 bytes:
    fields = struct.unpack("<4siiQQ", lead_in)
    # Keep track of information about file in fileinfo
    fileinfo = dict(zip(FILEINFO_NAMES, fields))
    fileinfo["decimated"] = not bool(fileinfo["toc"] & DECIMATE_MASK)
    # Make offsets relative to beginning of file:
    fileinfo["next_segment_offset"] += LEAD_IN_LENGTH
    fileinfo["raw_data_offset"] += LEAD_IN_LENGTH
    fileinfo["file_size"] = os.path.getsize(tdms_file.name)
    # Make sure next segment does not go beyond file capacity
    if fileinfo["next_segment_offset"] > fileinfo["file_size"]:
        fileinfo["next_segment_offset"] = fileinfo["file_size"]
    # navigate pointer to immediately after lead in data
    tdms_file.seek(LEAD_IN_LENGTH, 0)
    # Read number of channels
    n_channels = struct.unpack("i", tdms_file.read(4))[0] - 2
    fileinfo["n_channels"] = n_channels
    # Read length of object path:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    # skip over object path and raw data index:
    tdms_file.seek(var + 4, 1)
    # Read number of properties in this group:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    # loop through and read each property
    out = [_read_attr(tdms_file) for _ in range(var)]
    # Returns a pandas dataframe that we convert to dictionary
    out = dict(out)
    # Add other attributes not yet included
    out["n_channels"] = n_channels
    out["data_type"] = "strain_rate"
    out["data_units"] = ""
    out["dims"] = "time,distance"
    out["instrument_id"] = out.get("SystemInfomation.Devices0.SerialNum")
    # Rename some attributes to preferred names
    d_coord = _get_distance_coord(out)
    fileinfo["end_of_properties_offset"] = tdms_file.tell()
    tdms_file.seek(fileinfo["end_of_properties_offset"], 0)
    # skip over Group Information:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    tdms_file.seek(var + 8, 1)
    # skip over first channel path and length of index information:
    var = struct.unpack("<i", tdms_file.read(4))[0]
    tdms_file.seek(var + 4, 1)
    fileinfo["data_type"] = TDS_DATA_TYPE.get(struct.unpack("<i", tdms_file.read(4))[0])
    if fileinfo["data_type"] not in ("int16", "float32"):
        raise Exception("Unsupported TDMS data type: " + fileinfo["data_type"])
    # get number of samples by dividing amount of unread data by the
    # size of data per channel
    numofsamples = (
        (fileinfo["file_size"] - fileinfo["raw_data_offset"])
        / n_channels
        / np.dtype(fileinfo["data_type"]).itemsize
    )
    t_coord = _get_time_coord(out, numofsamples)
    out["coords"] = {"time": t_coord, "distance": d_coord}
    out.update(t_coord.get_attrs_dict("time"))
    out.update(d_coord.get_attrs_dict("distance"))
    return out, fileinfo


def _get_fileinfo(tdms_file, LEAD_IN_LENGTH=28):
    """Get info about file not included in the attributes."""
    attrs, fileinfo = _get_all_attrs(tdms_file)
    # Read Dimension of the raw data array (has to be 1):
    _ = struct.unpack("<i", tdms_file.read(4))[0]
    fileinfo["chunk_size"] = struct.unpack("<i", tdms_file.read(4))[0]
    return fileinfo, attrs


def _get_data(tdms_file, LEAD_IN_LENGTH=28):
    """Get all the data saved in the current file."""

    def get_segment_data(fileinfo, nch, dmap, nso, rdo):
        # seg1_length: length of recording indicated as raw_data in metadata for
        # each channel in bytes
        seg_length = int((nso - rdo) / nch / np.dtype(fileinfo["data_type"]).itemsize)
        channel_length = seg_length

        if fileinfo["decimated"]:
            # number of completely full chunks
            n_complete_blk = int(seg_length / fileinfo["chunk_size"])
            ax_ord = "C"
        else:
            n_complete_blk = 0
            ax_ord = "F"
        # use data from mapped file to fill variable raw_data
        raw_data = np.ndarray(
            (n_complete_blk, nch, fileinfo["chunk_size"]),
            dtype=fileinfo["data_type"],
            buffer=dmap,
            offset=rdo,
        )
        # Rotate the axes to [chunk_size, nblk, nch]
        raw_data = np.rollaxis(raw_data, 2)
        data_node = np.reshape(raw_data, (n_complete_blk * fileinfo["chunk_size"], nch))
        if n_complete_blk != seg_length / fileinfo["chunk_size"]:
            # If the last chunk isn't full there is some data left
            additional_samples = int(
                seg_length - n_complete_blk * fileinfo["chunk_size"]
            )
            additional_samples_offset = (
                rdo
                + n_complete_blk
                * nch
                * fileinfo["chunk_size"]
                * np.dtype(fileinfo["data_type"]).itemsize
            )
            raw_last_chunk = np.ndarray(
                (nch, additional_samples),
                dtype=fileinfo["data_type"],
                buffer=dmap,
                offset=additional_samples_offset,
                order=ax_ord,
            )
            # Rotate the axes to [samples, nch]
            raw_last_chunk = np.rollaxis(raw_last_chunk, 1)
            data_node = np.append(data_node, raw_last_chunk, axis=0)
            return data_node, channel_length

    fileinfo, attrs = _get_fileinfo(tdms_file)

    # map file contents to a variable dmap
    dmap = mmap.mmap(tdms_file.fileno(), 0, access=mmap.ACCESS_READ)
    # rdo: the start of the data in the file
    rdo = int(fileinfo["raw_data_offset"])
    # nch: number of channels
    nch = int(fileinfo["n_channels"])

    # nso: the beginning of the segment that comes next after the data
    nso = fileinfo["next_segment_offset"]
    # seg1_length: length of recording indicated as raw_data in metadata for
    # each channel in bytes

    flag = 0
    while flag == 0:
        cdata_node, cchannel_length = get_segment_data(fileinfo, nch, dmap, nso, rdo)
        if fileinfo["file_size"] == nso:
            flag = 1
        else:
            tdms_file.seek(nso + 12, 0)
            (next_seg_nso, next_seg_rdo) = struct.unpack("<qq", tdms_file.read(2 * 8))
            rdo = min(fileinfo["file_size"], nso + LEAD_IN_LENGTH + next_seg_rdo)
            nso = min(fileinfo["file_size"], nso + LEAD_IN_LENGTH + next_seg_nso)
        try:
            data_node = np.append(data_node, cdata_node, axis=1)
            channel_length += cchannel_length
        except NameError:
            data_node = cdata_node
            channel_length = cchannel_length

    return data_node, channel_length, attrs



"""Utilities for segy."""

from __future__ import annotations

import datetime

import numpy as np
from segyio import TraceField

import dascore as dc
from dascore.core import get_coord_manager

# --- Getting format/version


def _get_filtered_data_and_coords(segy_fi, coords, time=None, channel=None):
    """
    Read data from segy_file, possibly reading only subsections.

    Return filtered data and update coord manager.
    """
    traces_inds_to_read = np.arange(len(segy_fi.header), dtype=np.int64)
    time_slice = slice(None, None)
    traces = segy_fi.trace

    # filter time
    if time is not None:
        time_coord = coords.coord_map["time"]
        new_coord, time_slice = time_coord.select(time)
        coords = coords.update_coords(time=new_coord)

    # filter channel
    if channel:
        channel_coord = coords.coord_map["channel"]
        new_coord, channel_inds = channel_coord.select(channel)
        coords = coords.update_coords(channel=new_coord)
        traces_inds_to_read = traces_inds_to_read[channel_inds]

    # filter channels
    data_list = [traces[x][time_slice] for x in traces_inds_to_read]
    return np.stack(data_list, axis=-1), coords


def _get_coords(fi):
    """
    Get coordinates of the segy file.

    Time comes from the SEGY format of year, julian day, hour, minute, second.

    Distance axis is channel number. If the user knows the delta_x,
    then the axis should be modified.

    If a user knows the dx, change from channel to distance using
    patch.update_coords after reading
    """
    header_0 = fi.header[0]

    # get time array from SEGY headers
    starttime = _get_time_from_header(header_0)
    dt = dc.to_timedelta64(header_0[TraceField.TRACE_SAMPLE_INTERVAL] / 1000)
    ns = header_0[TraceField.TRACE_SAMPLE_COUNT]
    time_array = starttime + dt * np.arange(ns)

    # Get distance array from SEGY header
    channel = np.arange(len(fi.header))

    coords = get_coord_manager(
        {"time": time_array, "channel": channel}, dims=("time", "channel")
    )
    return coords


def _get_attrs(fi, coords, path, file_io):
    """Create Patch Attribute from SEGY header contents."""
    attrs = dc.PatchAttrs(
        path=path,
        file_version=file_io.version,
        file_format=file_io.name,
        coords=coords,
    )
    return attrs


def _get_time_from_header(header):
    """Creates a datetime64 object from SEGY header date information."""
    year = header[TraceField.YearDataRecorded]
    julday = header[TraceField.DayOfYear]
    hour = header[TraceField.HourOfDay]
    minute = header[TraceField.MinuteOfHour]
    second = header[TraceField.SecondOfMinute]
    # make those timedate64
    fmt = "%Y.%j.%H.%M.%S"
    s = f"{year}.{julday}.{hour}.{minute}.{second}"
    time = datetime.datetime.strptime(s, fmt)
    return dc.to_datetime64(time)



"""
Compatibility module for DASCore.

All components/functions that may be exchanged for other numpy/scipy
compatible libraries should go in this model.
"""
from __future__ import annotations

from contextlib import suppress

from numpy import floor, interp  # NOQA
from scipy.interpolate import interp1d  # NOQA
from scipy.ndimage import zoom  # NOQA
from scipy.signal import decimate, resample, resample_poly  # NOQA


class DataArray:
    """A dummy class for when xarray isn't installed."""


with suppress(ImportError):
    from xarray import DataArray  # NOQA


def array(array):
    """Wrapper function for creating 'immutable' arrays."""
    array.setflags(write=False)
    return array



"""DASDAE format utilities."""
from __future__ import annotations

import numpy as np
from tables import NodeError

import dascore as dc
from dascore.core.attrs import PatchAttrs
from dascore.core.coordmanager import get_coord_manager
from dascore.core.coords import get_coord
from dascore.utils.misc import suppress_warnings
from dascore.utils.patch import get_default_patch_name
from dascore.utils.time import to_int

# --- Functions for writing DASDAE format


def _create_or_get_group(h5, group, name):
    """Create a new group or get existing."""
    try:
        group = h5.create_group(group, name)
    except NodeError:
        group = getattr(group, name)
    return group


def _create_or_squash_array(h5, group, name, data):
    """Create a new array, if it exists delete and re-create."""
    try:
        array = h5.create_array(group, name, data)
    except NodeError:
        old_node = getattr(group, name)
        h5.remove_node(old_node)
        array = h5.create_array(group, name, data)
    return array


def _write_meta(hfile, file_version):
    """Write metadata to hdf5 file."""
    attrs = hfile.root._v_attrs
    attrs["__format__"] = "DASDAE"
    attrs["__DASDAE_version__"] = file_version
    attrs["__dascore__version__"] = dc.__version__


def _save_attrs_and_dims(patch, patch_group):
    """Save the attributes."""
    # copy attrs to group attrs
    # TODO will need to test if objects are serializable
    attr_dict = patch.attrs.model_dump(exclude_unset=True)
    for i, v in attr_dict.items():
        patch_group._v_attrs[f"_attrs_{i}"] = v
    patch_group._v_attrs["_dims"] = ",".join(patch.dims)


def _save_array(data, name, group, h5):
    """Save an array to a group, handle datetime flubbery."""
    # handle datetime conversions
    is_dt = np.issubdtype(data.dtype, np.datetime64)
    is_td = np.issubdtype(data.dtype, np.timedelta64)
    if is_dt or is_td:
        data = to_int(data)
    array_node = _create_or_squash_array(h5, group, name, data)
    array_node._v_attrs["is_datetime64"] = is_dt
    array_node._v_attrs["is_timedelta64"] = is_td


def _save_coords(patch, patch_group, h5):
    """Save coordinates."""
    cm = patch.coords
    for name, coord in cm.coord_map.items():
        dims = cm.dim_map[name]
        # First save coordinate arrays
        data = coord.values
        save_name = f"_coord_{name}"
        _save_array(data, save_name, patch_group, h5)
        # then save dimensions of coordinates
        save_name = f"_cdims_{name}"
        patch_group._v_attrs[save_name] = ",".join(dims)


def _save_patch(patch, wave_group, h5):
    """Save the patch to disk."""
    name = get_default_patch_name(patch)
    patch_group = _create_or_get_group(h5, wave_group, name)
    _save_attrs_and_dims(patch, patch_group)
    _save_coords(patch, patch_group, h5)
    # add data
    if patch.data.shape:
        _create_or_squash_array(h5, patch_group, "data", patch.data)


# --- Functions for reading


def _get_attrs(patch_group):
    """Get the saved attributes form the group attrs."""
    out = {}
    attrs = [x for x in patch_group._v_attrs._f_list() if x.startswith("_attrs_")]
    for attr_name in attrs:
        key = attr_name.replace("_attrs_", "")
        val = patch_group._v_attrs[attr_name]
        # need to unpack one value arrays
        if isinstance(val, np.ndarray) and not val.shape:
            val = np.array([val])[0]
        out[key] = val
    with suppress_warnings(DeprecationWarning):
        return PatchAttrs(**out)


def _read_array(table_array):
    """Read an array into numpy."""
    data = table_array[:]
    if table_array._v_attrs["is_datetime64"]:
        data = data.view("datetime64[ns]")
    if table_array._v_attrs["is_timedelta64"]:
        data = data.view("timedelta64[ns]")
    return data


def _get_coords(patch_group, dims, attrs2):
    """Get the coordinates from a patch group."""
    coord_dict = {}  # just store coordinates here
    coord_dim_dict = {}  # stores {coord_name: ((dims, ...), coord)}
    for coord in [x for x in patch_group if x.name.startswith("_coord_")]:
        name = coord.name.replace("_coord_", "")
        array = _read_array(coord)
        coord = get_coord(
            values=array,
            units=getattr(attrs2, f"{name}_units", None),
            step=getattr(attrs2, f"{name}_step", None),
        )
        coord_dict[name] = coord
    # associates coordinates with dimensions
    c_dims = [x for x in patch_group._v_attrs._f_list() if x.startswith("_cdims")]
    for coord_name in c_dims:
        name = coord_name.replace("_cdims_", "")
        value = patch_group._v_attrs[coord_name]
        assert name in coord_dict, "Should already have loaded coordinate array"
        coord_dim_dict[name] = (tuple(value.split(".")), coord_dict[name])
        # add dimensions to coordinates that have them.
    cm = get_coord_manager(coord_dim_dict, dims=dims)
    return cm


def _get_dims(patch_group):
    """Get the dims tuple from the patch group."""
    dims = patch_group._v_attrs["_dims"]
    if not dims:
        out = ()
    else:
        out = tuple(dims.split(","))
    return out


def _read_patch(patch_group, **kwargs):
    """Read a patch group, return Patch."""
    attrs = _get_attrs(patch_group)
    dims = _get_dims(patch_group)
    coords = _get_coords(patch_group, dims, attrs)
    # Note, previously this was wrapped with try, except (Index, KeyError)
    # and the data = np.array(None) in except block. Not sure, why, removed
    # try except.
    if kwargs:
        coords, data = coords.select(array=patch_group["data"], **kwargs)
    else:
        data = patch_group["data"][:]
    return dc.Patch(data=data, coords=coords, dims=dims, attrs=attrs)


def _get_contents_from_patch_groups(h5, file_version, file_format="DASDAE"):
    """Get the contents from each patch group."""
    out = []
    for group in h5.iter_nodes("/waveforms"):
        contents = _get_patch_content_from_group(group)
        # populate file info
        contents["file_version"] = file_version
        contents["file_format"] = file_format
        contents["path"] = h5.filename
        # suppressing warnings because old dasdae files will issue warning
        # due to d_dim rather than dim_step. TODO fix test files in the future
        with suppress_warnings(DeprecationWarning):
            out.append(dc.PatchAttrs(**contents))
    return out


def _get_patch_content_from_group(group):
    """Get patch content from a single node."""
    attrs = group._v_attrs
    out = {}
    for key in attrs._f_list():
        value = getattr(attrs, key)
        new_key = key.replace("_attrs_", "")
        # need to unpack 0 dim arrays.
        if isinstance(value, np.ndarray) and not value.shape:
            value = np.atleast_1d(value)[0]
        out[new_key] = value
    # rename dims
    out["dims"] = out.pop("_dims")
    return out



"""Module for performing differentiation on patches."""
from __future__ import annotations

from operator import truediv

import numpy as np

from dascore.constants import PatchType
from dascore.utils.misc import iterate, optional_import
from dascore.utils.patch import (
    _get_data_units_from_dims,
    _get_dx_or_spacing_and_axes,
    patch_function,
)


def _grad_diff(ar, axes, dxs):
    """Perform differentiation on the data using np.grad."""
    # iteratively apply diff along each axis.
    for ax, dx in zip(axes, dxs):
        ar = np.gradient(ar, dx, axis=ax, edge_order=2)
    return ar


def _findiff_diff(data, axis, order, dx):
    """Use findiff to get differences along specified dimensions."""
    # If more than one axis, need to make tuples.
    findiff = optional_import("findiff")
    if len(axis) > 1:
        orders = [1] * len(axis)  # always first order (first derivative).
        tuples = [tuple(x) for x in zip(axis, dx, orders)]
        func = findiff.FinDiff(*tuples, acc=order)
    else:
        func = findiff.FinDiff(axis[0], dx[0], 1, acc=order)
    out = func(data)
    return out


@patch_function()
def differentiate(
    patch: PatchType,
    dim: str | None,
    order=2,
) -> PatchType:
    """
    Calculate first derivative along dimension(s) using centeral diferences.

    The shape of the output patch is the same as the input patch. Derivative
    along edges are calculated with the same order (accuarcy) as centeral
    points using non-centered stencils.

    Parameters
    ----------
    patch
        The patch to differentiate.
    dim
        The dimension(s) along which to differentiate. If None differentiates
        over all dimensions.
    order
        The order of the differentiation operator. Must be a possitive, even
        integar.

    Notes
    -----
    For order=2 (the default) numpy's gradient function is used. When
    order != the optional package findiff must be installed in which case
    order is interpreted as accuracy ("order" means order of differention
    in that package).

    Examples
    --------
    >>> import dascore as dc
    >>> patch = dc.get_example_patch()
    >>> # 2nd order stencil for 1st derivative along time dimension
    >>> patch_diff_1 = patch.differentiate(dim='time', order=2)
    >>> # 1st derivative along all dimensions
    >>> patch_diff_2 = patch.differentiate(dim=None)
    """
    dims = iterate(dim if dim is not None else patch.dims)
    dx_or_spacing, axes = _get_dx_or_spacing_and_axes(patch, dims)
    if order == 2:
        new_data = _grad_diff(patch.data, axes, dx_or_spacing)
    else:
        new_data = _findiff_diff(patch.data, axes, order, dx_or_spacing)
    # update units
    data_units = _get_data_units_from_dims(patch, dims, truediv)
    attrs = patch.attrs.update(data_units=data_units)
    return patch.new(data=new_data, attrs=attrs)
